from __future__ import annotations

import textwrap
import types
import typing as _t
from dataclasses import dataclass
from enum import Enum

__all__ = [
    "Element",
    "Terminal",
    "NonTerminal",
    "Comment",
    "Sequence",
    "Stack",
    "NoBreak",
    "Choice",
    "Optional",
    "OneOrMore",
    "ZeroOrMore",
    "Barrier",
    "Group",
    "LineBreak",
    "skip",
    "terminal",
    "non_terminal",
    "comment",
    "sequence",
    "stack",
    "no_break",
    "choice",
    "optional",
    "one_or_more",
    "zero_or_more",
    "barrier",
    "group",
]

T = _t.TypeVar("T")


@dataclass(frozen=True, kw_only=True)
class _Annotation:
    name: str | None = None
    description: str | None = None


class Terminal(_t.TypedDict, _t.Generic[T], total=False):
    terminal: _t.Annotated[
        _t.Required[str],
        _Annotation(
            description="Describes a terminal node with optional additional settings.",
        ),
    ]

    href: _t.Annotated[
        str | None,
        _Annotation(
            description="Makes text node into a hyperlink.",
        ),
    ]

    title: _t.Annotated[
        str | None,
        _Annotation(
            description="Title for hyperlink.",
        ),
    ]

    css_class: _t.Annotated[
        str | None,
        _Annotation(
            description="Adds CSS class to node's ``<g>`` element.",
        ),
    ]

    resolve: _t.Annotated[
        bool | None,
        _Annotation(
            description="If set to `False`, this node will not be passed to `HrefResolver`",
        ),
    ]

    resolver_data: _t.Annotated[
        T | None,
        _Annotation(
            description="Additional data for `HrefResolver`.",
        ),
    ]


class NonTerminal(_t.TypedDict, _t.Generic[T], total=False):
    non_terminal: _t.Annotated[
        _t.Required[str],
        _Annotation(
            description="Describes a non-terminal node with optional additional settings.",
        ),
    ]

    href: _t.Annotated[
        str | None,
        _Annotation(
            description="Makes text node into a hyperlink.",
        ),
    ]

    title: _t.Annotated[
        str | None,
        _Annotation(
            description="Title for hyperlink.",
        ),
    ]

    css_class: _t.Annotated[
        str | None,
        _Annotation(
            description="Adds CSS class to node's ``<g>`` element.",
        ),
    ]

    resolve: _t.Annotated[
        bool | None,
        _Annotation(
            description="If set to `False`, this node will not be passed to `HrefResolver`",
        ),
    ]

    resolver_data: _t.Annotated[
        T | None,
        _Annotation(
            description="Additional data for `HrefResolver`.",
        ),
    ]


class Comment(_t.TypedDict, _t.Generic[T], total=False):
    comment: _t.Annotated[
        _t.Required[str],
        _Annotation(
            description="Describes a comment node with optional additional settings.",
        ),
    ]

    href: _t.Annotated[
        str | None,
        _Annotation(
            description="Makes text node into a hyperlink.",
        ),
    ]

    title: _t.Annotated[
        str | None,
        _Annotation(
            description="Title for hyperlink.",
        ),
    ]

    css_class: _t.Annotated[
        str | None,
        _Annotation(
            description="Adds CSS class to node's ``<g>`` element.",
        ),
    ]

    resolve: _t.Annotated[
        bool | None,
        _Annotation(
            description="If set to `False`, this node will not be passed to `HrefResolver`",
        ),
    ]

    resolver_data: _t.Annotated[
        T | None,
        _Annotation(
            description="Additional data for `HrefResolver`.",
        ),
    ]


class Sequence(_t.TypedDict, _t.Generic[T], total=False):
    sequence: _t.Annotated[
        _t.Required[list[Element[T]]],
        _Annotation(
            description="Describes an automatically wrapped sequence of elements.",
        ),
    ]

    linebreaks: _t.Annotated[
        (
            _t.Literal["HARD", "SOFT", "DEFAULT", "NO_BREAK"]
            | LineBreak
            | list[_t.Literal["HARD", "SOFT", "DEFAULT", "NO_BREAK"] | LineBreak]
        )
        | None,
        _Annotation(
            description="Specifies line breaks after each element of the sequence.",
        ),
    ]


class Stack(_t.TypedDict, _t.Generic[T], total=False):
    stack: _t.Annotated[
        _t.Required[list[Element[T]]],
        _Annotation(
            description="Describes a sequence of elements that wraps after each element.",
        ),
    ]


class NoBreak(_t.TypedDict, _t.Generic[T], total=False):
    no_break: _t.Annotated[
        _t.Required[list[Element[T]]],
        _Annotation(
            description="Describes a sequence of elements that doesn't wrap.",
        ),
    ]


class Choice(_t.TypedDict, _t.Generic[T], total=False):
    choice: _t.Annotated[
        _t.Required[list[Element[T]]],
        _Annotation(
            description="Describes a choice between several elements.",
        ),
    ]

    default: _t.Annotated[
        int | None,
        _Annotation(
            description="Index of item that will be placed on the main line.",
        ),
    ]


class Optional(_t.TypedDict, _t.Generic[T], total=False):
    optional: _t.Annotated[
        _t.Required[Element[T]],
        _Annotation(
            description="Describes an optional element.",
        ),
    ]

    skip: _t.Annotated[
        bool | None,
        _Annotation(
            description="If set to `True`, the optional element will be rendered off the main line.",
        ),
    ]

    skip_bottom: _t.Annotated[
        bool | None,
        _Annotation(
            description="If set to `True`, the skip line will be rendered below the skipped element.",
        ),
    ]


class OneOrMore(_t.TypedDict, _t.Generic[T], total=False):
    one_or_more: _t.Annotated[
        _t.Required[Element[T]],
        _Annotation(
            description="Describes a repeated element.",
        ),
    ]

    repeat: _t.Annotated[
        Element[T] | None,
        _Annotation(
            description="An element that will be placed on the backwards path.",
        ),
    ]

    # repeat_top: _t.Annotated[
    #     bool | None,
    #     _Annotation(
    #         description="If set to `True`, the repeat element will be rendered above the repeated one.",
    #     ),
    # ]


class ZeroOrMore(_t.TypedDict, _t.Generic[T], total=False):
    zero_or_more: _t.Annotated[
        _t.Required[Element[T]],
        _Annotation(
            description="Describes an optional repeated element.",
        ),
    ]

    repeat: _t.Annotated[
        Element[T] | None,
        _Annotation(
            description="An element that will be placed on the backwards path.",
        ),
    ]

    # repeat_top: _t.Annotated[
    #     bool | None,
    #     _Annotation(
    #         description="If set to `True`, the repeat element will be rendered above the repeated one.",
    #     ),
    # ]

    skip: _t.Annotated[
        bool | None,
        _Annotation(
            description="If set to `True`, the optional repeated element will be rendered off the main line.",
        ),
    ]

    skip_bottom: _t.Annotated[
        bool | None,
        _Annotation(
            description="If set to `True`, the skip line will be rendered below the skipped element.",
        ),
    ]


class Barrier(_t.TypedDict, _t.Generic[T], total=False):
    barrier: _t.Annotated[
        _t.Required[Element[T]],
        _Annotation(
            description=(
                "Isolates an element and disables optimizations that merge "
                "lines between this element and the rest of the diagram."
            )
        ),
    ]


class Group(_t.TypedDict, _t.Generic[T], total=False):
    group: _t.Annotated[
        _t.Required[Element[T]],
        _Annotation(
            description="Draws a box around some element.",
        ),
    ]

    text: _t.Annotated[
        str | None,
        _Annotation(
            description="Group's text.",
        ),
    ]

    href: _t.Annotated[
        str | None,
        _Annotation(
            description="Makes text into a hyperlink.",
        ),
    ]

    title: _t.Annotated[
        str | None,
        _Annotation(
            description="Title for hyperlink.",
        ),
    ]

    css_class: _t.Annotated[
        str | None,
        _Annotation(
            description="Adds CSS class to group's ``<g>`` element.",
        ),
    ]


Element: _t.TypeAlias = _t.Annotated[
    None
    | str
    | list["Element[T]"]
    | Terminal[T]
    | NonTerminal[T]
    | Comment[T]
    | Sequence[T]
    | Stack[T]
    | NoBreak[T]
    | Choice[T]
    | Optional[T]
    | OneOrMore[T]
    | ZeroOrMore[T]
    | Barrier[T]
    | Group[T],
    _Annotation(name="Element"),
]


class LineBreak(Enum):
    """
    Type of a line break that can be specified when creating a sequence.

    """

    HARD = "HARD"
    """
    Always breaks a line in this position.

    """

    SOFT = "SOFT"
    """
    Breaks a line in this position if the sequence can't fit
    in allowed dimensions.

    """

    DEFAULT = "DEFAULT"
    """
    Prefers not to break line in this position unless necessary.

    """

    NO_BREAK = "NO_BREAK"
    """
    Disables breaking in this position.

    """


def skip() -> Element[T]:
    """
    Create an element that renders as a single line without content.

    See `Element` for more information.

    """

    return None


def terminal(
    text: str,
    *,
    href: str | None = None,
    title: str | None = None,
    css_class: str | None = None,
    resolve: bool | None = None,
    resolver_data: T | None = None,
) -> Element[T]:
    """
    Create a terminal node with optional additional settings.

    See `Terminal` for description of parameters.

    """
    return {
        "terminal": text,
        "href": href,
        "title": title,
        "css_class": css_class,
        "resolve": resolve,
        "resolver_data": resolver_data,
    }


def non_terminal(
    text: str,
    *,
    href: str | None = None,
    title: str | None = None,
    css_class: str | None = None,
    resolve: bool | None = None,
    resolver_data: T | None = None,
) -> Element[T]:
    """
    Create a non-terminal node with optional additional settings.

    See `NonTerminal` for description of parameters.

    """

    return {
        "non_terminal": text,
        "href": href,
        "title": title,
        "css_class": css_class,
        "resolve": resolve,
        "resolver_data": resolver_data,
    }


def comment(
    text: str,
    *,
    href: str | None = None,
    title: str | None = None,
    css_class: str | None = None,
    resolve: bool | None = None,
    resolver_data: T | None = None,
) -> Element[T]:
    """
    Create a comment node with optional additional settings.

    See `Comment` for description of parameters.

    """

    return {
        "comment": text,
        "href": href,
        "title": title,
        "css_class": css_class,
        "resolve": resolve,
        "resolver_data": resolver_data,
    }


def sequence(
    *items: Element[T],
    linebreaks: (
        _t.Literal["HARD", "SOFT", "DEFAULT", "NO_BREAK"]
        | LineBreak
        | list[_t.Literal["HARD", "SOFT", "DEFAULT", "NO_BREAK"] | LineBreak]
        | None
    ) = None,
) -> Element[T]:
    """
    Create an automatically wrapped sequence of elements.

    See `Sequence` for description of parameters.

    """

    return {
        "sequence": list(items),
        "linebreaks": linebreaks,
    }


def stack(*items: Element[T]) -> Element[T]:
    """
    Create a sequence of elements that wraps after each element.

    See `Stack` for description of parameters.

    """

    return {
        "stack": list(items),
    }


def no_break(*items: Element[T]) -> Element[T]:
    """
    Create a sequence of elements that doesn't wrap.

    See `NoBreak` for description of parameters.

    """

    return {
        "no_break": list(items),
    }


def choice(*items: Element[T], default: int = 0) -> Element[T]:
    """
    Create a choice between several elements.

    See `Choice` for description of parameters.

    """

    return {
        "choice": list(items),
        "default": default,
    }


def optional(
    *items: Element[T], skip: bool = False, skip_bottom: bool = False
) -> Element[T]:
    """
    Create an optional element.

    See `Optional` for description of parameters.

    """

    return {
        "optional": list(items),
        "skip": skip,
        "skip_bottom": skip_bottom,
    }


def one_or_more(
    *items: Element[T],
    repeat: Element[T] | None = None,
    # repeat_top: bool = False,
) -> Element[T]:
    """
    Create a repeated element.

    See `OneOrMore` for description of parameters.

    """

    return {
        "one_or_more": list(items),
        "repeat": repeat,
        # "repeat_top": repeat_top,
    }


def zero_or_more(
    *items: Element[T],
    repeat: Element[T] | None = None,
    # repeat_top: bool = False,
    skip: bool = False,
    skip_bottom: bool = False,
) -> Element[T]:
    """
    Create an optional repeated element.

    See `ZeroOrMore` for description of parameters.

    """

    return {
        "zero_or_more": list(items),
        "repeat": repeat,
        # "repeat_top": repeat_top,
        "skip": skip,
        "skip_bottom": skip_bottom,
    }


def barrier(
    *items: Element[T],
) -> Element[T]:
    """
    Create a barrier element.

    See `Barrier` for description of parameters.

    """

    return {
        "barrier": list(items),
    }


def group(
    *items: Element[T],
    text: str | None = None,
    href: str | None = None,
    title: str | None = None,
    css_class: str | None = None,
) -> Element[T]:
    """
    Create a group element.

    See `Group` for description of parameters.

    """

    return {
        "group": list(items),
        "text": text,
        "href": href,
        "title": title,
        "css_class": css_class,
    }


def _to_json_schema(
    cls,
    globals,
    locals,
) -> dict[str, _t.Any]:
    # Based on https://github.com/Udzu/dataglasses

    defs: dict[str, _t.Any] = {}
    saved: dict[_t.Any, str] = {}

    def _json_schema(
        cls: _t.Any,
    ) -> dict[str, _t.Any]:
        basic_types = {
            bool: "boolean",
            int: "integer",
            float: "number",
            str: "string",
            types.NoneType: "null",
        }

        if cls is None:
            cls = types.NoneType

        if cls in basic_types:
            return {"type": basic_types[cls]}

        if cls in saved:
            return {"$ref": f"#/$defs/{saved[cls]}"}

        origin = _t.get_origin(cls)

        if _t.is_typeddict(origin):
            if cls.__qualname__ not in defs:
                defn: dict[str, _t.Any] = {
                    "type": "object",
                    "properties": {},
                    "required": [],
                    "additionalProperties": False,
                    "description": textwrap.dedent(
                        origin.__doc__ or cls.__qualname__
                    ).strip(),
                }
                defs[cls.__qualname__] = defn

                for name, ann in _t.get_type_hints(
                    origin, globals, locals, include_extras=True
                ).items():
                    if _t.get_origin(ann) is _t.Required:
                        defn["required"].append(name)
                    defn["properties"][name] = _json_schema(ann)

            return {"$ref": f"#/$defs/{cls.__qualname__}"}

        if isinstance(cls, (str, _t.ForwardRef)):
            ref = _t.ForwardRef(cls) if isinstance(cls, str) else cls
            evaluated_type = ref._evaluate(globals, locals, recursive_guard=frozenset())
            return _json_schema(evaluated_type)
        elif origin is list:
            sequence_type = _t.get_args(cls)[0]
            return {"type": "array", "items": _json_schema(sequence_type)}
        elif origin is dict:
            key_type, value_type = _t.get_args(cls)
            if key_type is not str:
                raise ValueError(f"Unsupported non-string mapping key type: {key_type}")
            return {
                "type": "object",
                "patternProperties": {"^.*$": _json_schema(value_type)},
            }
        elif origin in (_t.Union, types.UnionType, _t.Required, _t.NotRequired):
            union_types = _t.get_args(cls)
            return {"anyOf": [_json_schema(t) for t in union_types]}
        elif origin is tuple:
            tuple_types = _t.get_args(cls)
            if len(tuple_types) == 2 and tuple_types[1] == Ellipsis:
                return {
                    "type": "array",
                    "items": _json_schema(tuple_types[0]),
                }
            else:
                return {
                    "type": "array",
                    "prefixItems": [_json_schema(t) for t in tuple_types],
                    "minItems": len(tuple_types),
                    "maxItems": len(tuple_types),
                }
        elif origin == _t.Literal:
            return {"enum": list(_t.get_args(cls))}
        elif origin == _t.Annotated:
            annotated_type, data = _t.get_args(cls)
            if not isinstance(data, _Annotation):
                return _json_schema(annotated_type)
            if data.name:
                saved[cls] = data.name

            defn = _json_schema(annotated_type)
            defn["description"] = data.description

            if data.name:
                defs[data.name] = defn
                return {"$ref": f"#/$defs/{data.name}"}
            else:
                return defn
        elif isinstance(cls, type) and issubclass(cls, Enum):
            return {"enum": [e.value for e in cls]}
        elif isinstance(cls, _t.TypeVar):
            return {}
        else:
            raise ValueError(f"Unsupported type {cls}")

    schema: dict[str, _t.Any] = {}
    schema["$schema"] = "https://json-schema.org/draft/2020-12/schema"
    schema.update(_json_schema(cls))
    schema["$defs"] = defs
    return schema


def _main():
    import argparse
    import json
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument("-o", type=argparse.FileType("w"), default=sys.stdout)
    args = parser.parse_args()

    print(
        json.dumps(_to_json_schema(Element, globals(), locals()), indent=2), file=args.o
    )


if __name__ == "__main__":
    _main()



from __future__ import annotations

import base64
import io
import math
import re
import sys
import typing as _t
import uuid
from dataclasses import dataclass, field

from syntax_diagrams._impl.render import (
    ConnectionDirection,
    ConnectionType,
    Direction,
    LayoutContext,
    LayoutSettings,
    Line,
    NodeStyle,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.tree import Element
from syntax_diagrams._impl.tree.barrier import Barrier
from syntax_diagrams._impl.tree.end import End
from syntax_diagrams._impl.tree.sequence import Sequence
from syntax_diagrams._impl.vec import Vec
from syntax_diagrams.element import LineBreak
from syntax_diagrams.measure import TextMeasure
from syntax_diagrams.render import ArrowStyle, EndClass, SvgRenderSettings
from syntax_diagrams.resolver import HrefResolver

T = _t.TypeVar("T")


def render_svg(
    node: Element[T],
    /,
    *,
    max_width: int | None = None,
    settings: SvgRenderSettings = SvgRenderSettings(),
    reverse: bool | None = None,
    href_resolver: HrefResolver[T] = HrefResolver(),
    dump_debug_data: bool = False,
) -> str:
    if max_width is None:
        max_width = settings.max_width
    if reverse is None:
        reverse = settings.reverse

    max_width = max(0, max_width - settings.padding[3] - settings.padding[1])

    node = Sequence(
        [End(), Barrier(Sequence([node, End(reverse=True)], LineBreak.NO_BREAK))],
        LineBreak.NO_BREAK,
    )

    layout_settings = svg_layout_settings(settings)
    layout_settings.href_resolver = href_resolver
    node.calculate_layout(
        layout_settings,
        LayoutContext(
            width=max_width,
            is_outer=True,
            start_connection=ConnectionType.NULL,
            start_top_is_clear=True,
            start_bottom_is_clear=True,
            start_direction=ConnectionDirection.STRAIGHT,
            end_connection=ConnectionType.NULL,
            end_top_is_clear=True,
            end_bottom_is_clear=True,
            end_direction=ConnectionDirection.STRAIGHT,
            allow_shrinking_stacks=True,
        ),
    )

    render = SvgRender(
        (settings.padding[3] + node.display_width + settings.padding[1]),
        (
            settings.padding[0]
            + node.up
            + node.height
            + node.down
            + settings.padding[2]
            + 1
        ),
        layout_settings,
        settings.css_class,
        settings.css_style,
        settings.title,
        settings.description,
        settings.arrow_style,
        settings.arrow_length,
        settings.arrow_cross_length,
        dump_debug_data,
    )

    pos = Vec(settings.padding[3], settings.padding[0] + node.up)
    if not reverse:
        start_connection_pos = pos
        end_connection_pos = pos + Vec(node.width, node.height)
    else:
        start_connection_pos = pos + Vec(node.display_width, 0)
        end_connection_pos = pos + Vec(node.display_width - node.width, node.height)

    node.render(
        render,
        RenderContext(
            pos=start_connection_pos,
            start_connection_pos=start_connection_pos,
            end_connection_pos=end_connection_pos,
            reverse=reverse,
        ),
    )

    if dump_debug_data:
        return render.debug_data()
    else:
        return render.to_string()


def svg_layout_settings(settings: SvgRenderSettings = SvgRenderSettings()):
    return LayoutSettings(
        horizontal_seq_separation=settings.horizontal_seq_separation,
        vertical_choice_separation_outer=settings.vertical_choice_separation_outer,
        vertical_choice_separation=settings.vertical_choice_separation,
        vertical_seq_separation_outer=settings.vertical_seq_separation_outer,
        vertical_seq_separation=settings.vertical_seq_separation,
        arc_radius=settings.arc_radius,
        arc_margin=settings.arc_margin,
        terminal_text_measure=settings.terminal_text_measure,
        terminal_horizontal_padding=settings.terminal_horizontal_padding,
        terminal_radius=settings.terminal_radius,
        terminal_vertical_padding=settings.terminal_vertical_padding,
        non_terminal_text_measure=settings.non_terminal_text_measure,
        non_terminal_horizontal_padding=settings.non_terminal_horizontal_padding,
        non_terminal_radius=settings.non_terminal_radius,
        non_terminal_vertical_padding=settings.non_terminal_vertical_padding,
        comment_text_measure=settings.comment_text_measure,
        comment_horizontal_padding=settings.comment_horizontal_padding,
        comment_radius=settings.comment_radius,
        comment_vertical_padding=settings.comment_vertical_padding,
        group_text_measure=settings.group_text_measure,
        group_vertical_padding=settings.group_vertical_padding,
        group_horizontal_padding=settings.group_horizontal_padding,
        group_vertical_margin=settings.group_vertical_margin,
        group_horizontal_margin=settings.group_horizontal_margin,
        group_thickness=0,
        group_radius=settings.group_radius,
        group_text_vertical_offset=settings.group_text_vertical_offset,
        group_text_horizontal_offset=settings.group_text_horizontal_offset,
        marker_width=20,
        marker_projected_height=10,
        hidden_symbol_escape=("\0", "\0"),
        end_class=settings.end_class,
    )


class SvgRender(Render[T], _t.Generic[T]):
    def __init__(
        self,
        width: int,
        height: int,
        settings: LayoutSettings[T],
        css_class: str,
        css: str | dict[str, dict[str, str]] | None,
        title: str | None,
        description: str | None,
        arrow_style: ArrowStyle,
        arrow_length: int,
        arrow_cross_length: int,
        dump_debug_data: bool = False,
    ):
        super().__init__(settings, dump_debug_data)

        self._width = width

        self._root = SvgRender._SvgElement(
            "svg",
            {
                "xmlns": "http://www.w3.org/2000/svg",
                "xmlns:xlink": "http://www.w3.org/1999/xlink",
                "width": width,
                "height": height,
                "class": css_class,
                "viewBox": f"0 0 {width} {height}",
                "role": "img",
            },
        )

        if title is not None:
            self._root.attrs["aria-label"] = title
            self._root.elem("title").children.append(self._e(title))
        if description is not None:
            self._root.elem("desc").children.append(self._e(description))

        self._arrow_style = arrow_style
        self._arrow_length = arrow_length
        self._arrow_cross_length = arrow_cross_length
        self._arrow_id = (
            f"arrow-{base64.urlsafe_b64encode(uuid.uuid4().bytes).decode()}"
        )
        self._arrow_class = f"arrow arrow-{str(arrow_style.value).lower()}"

        defs = self._root.elem("defs")
        match self._arrow_style:
            case ArrowStyle.NONE:
                pass
            case ArrowStyle.TRIANGLE:
                defs.elem(
                    "path",
                    {
                        "id": self._arrow_id,
                        "class": self._arrow_class,
                        "d": f"M 0 0 L -{arrow_length} -{arrow_cross_length} L -{arrow_length} {arrow_cross_length} z",
                    },
                )
            case ArrowStyle.STEALTH:
                defs.elem(
                    "path",
                    {
                        "id": self._arrow_id,
                        "class": self._arrow_class,
                        "d": f"M 0 0 L -{arrow_length} -{arrow_cross_length} L -{3 * arrow_length / 4} 0 L -{arrow_length} {arrow_cross_length} z",
                    },
                )
            case ArrowStyle.BARB:
                defs.elem(
                    "path",
                    {
                        "id": self._arrow_id,
                        "class": self._arrow_class,
                        "d": f"M 0 0 L -{arrow_length} -{arrow_cross_length} M 0 0 L -{arrow_length} {arrow_cross_length}",
                    },
                )
            case ArrowStyle.HARPOON:
                defs.elem(
                    "path",
                    {
                        "id": self._arrow_id,
                        "class": self._arrow_class,
                        "d": f"M 0 0 L -{arrow_length} {arrow_cross_length} L -{3 * arrow_length / 4} 0 z",
                    },
                )
            case ArrowStyle.HARPOON_UP:
                defs.elem(
                    "path",
                    {
                        "id": self._arrow_id,
                        "class": self._arrow_class,
                        "d": f"M 0 0 L -{arrow_length} -{arrow_cross_length} L -{3 * arrow_length / 4} 0 z",
                    },
                )

        if css:
            if not isinstance(css, str):
                stream = io.StringIO()
                for rule, items in css.items():
                    stream.write(rule)
                    stream.write("{")
                    for name, value in items.items():
                        stream.write(name)
                        stream.write(":")
                        stream.write(value)
                        stream.write(";")
                    stream.write("}")
                css = stream.getvalue()

            self._style = self._root.elem("style")
            self._style.children.append(self._e(css))

        self._elems = [self._root.elem("g")]

    def write(self, f=sys.stdout):
        self._root.write_svg(f, self)
        f.flush()

    def to_string(self):
        stream = io.StringIO()
        self.write(stream)
        return stream.getvalue()

    def enter(self, node: Element[_t.Any]):
        super().enter(node)

        attrs = {
            "class": "elem",
        }

        if self._debug:
            attrs["data-dbg-id"] = self._make_debug_id(node)

        self._elems.append(self._elem.elem("g", attrs))

    def exit(self):
        super().exit()

        self._elems.pop()

    @property
    def _elem(self) -> SvgRender._SvgElement:
        return self._elems[-1]

    def line(self, pos: Vec, reverse: bool = False, css_class: str = "") -> Line:
        return SvgRender._SvgLine(self, pos, reverse, css_class)

    def node(
        self,
        pos: Vec,
        style: NodeStyle,
        css_class: str | None,
        content_width: int,
        up: int,
        down: int,
        radius: int,
        padding: int,
        text_width: int,
        text_height: int,
        text: str,
        href: str | None,
        title: str | None,
    ):
        if css_class:
            css_class += " node "
        else:
            css_class = "node "
        match style:
            case NodeStyle.TERMINAL:
                css_class += " terminal"
                measure = self.settings.terminal_text_measure
            case NodeStyle.NON_TERMINAL:
                css_class += " non-terminal"
                measure = self.settings.non_terminal_text_measure
            case NodeStyle.COMMENT:
                css_class += " comment"
                measure = self.settings.comment_text_measure
        g = self._elem.elem(
            "g",
            {
                "class": css_class,
            },
        )

        g.elem(
            "rect",
            {
                "x": pos.x,
                "y": pos.y - up,
                "width": content_width,
                "height": up + down,
                "rx": radius,
                "ry": radius,
            },
        )

        if href:
            g = g.elem(
                "a",
                {
                    "xlink:href": href,
                    "title": title,
                },
            )

        self._make_text(
            pos.x + content_width / 2, pos.y, text, text_height, measure
        ).add_to(g)

    def group(
        self,
        pos: Vec,
        width: int,
        height: int,
        css_class: str | None,
        text_width: int,
        text_height: int,
        text: str | None,
        href: str | None,
        title: str | None,
    ):
        if css_class:
            css_class += " group "
        else:
            css_class = "group "

        g = self._elem.elem(
            "g",
            {
                "class": css_class,
            },
        )

        g.elem(
            "rect",
            {
                "x": pos.x,
                "y": pos.y,
                "width": width,
                "height": height,
                "rx": self.settings.group_radius,
                "ry": self.settings.group_radius,
            },
        )

        if not text:
            return

        if href:
            g = g.elem(
                "a",
                {
                    "xlink:href": href,
                    "title": title,
                },
            )

        self._make_text(
            pos.x + self.settings.group_text_horizontal_offset,
            pos.y - self.settings.group_text_vertical_offset,
            text,
            text_height,
            self.settings.group_text_measure,
            vertical_center=False,
        ).add_to(g)

    _UNESCAPE_RE = re.compile(r"\0(.*?)\0")

    def _make_text(
        self,
        x: float,
        y: float,
        text: str,
        text_height: int,
        measure: TextMeasure,
        vertical_center: bool = True,
    ) -> _SvgElement:
        style = f"font-size: {measure.font_size}px"

        lines = text.splitlines()
        line_height = text_height / len(lines) if lines else 0
        if vertical_center:
            # Pain =(
            ascent = measure.ascent + (measure.line_height - measure.font_size) / 2
            text_offset = text_height / 2 - ascent
        else:
            text_offset = text_height - line_height / 2

        g = SvgRender._SvgElement("g", {}, [])
        for i, line in enumerate(lines):
            e = g.elem(
                "text",
                {"x": x, "y": y - text_offset + i * line_height, "style": style},
            )

            j = 0
            for esc in self._UNESCAPE_RE.finditer(line):
                if part := line[j : esc.start()]:
                    e.children.append(self._e(part))
                j = esc.end()
                e.elem("tspan", {"class": "escape"}, [self._e(esc.group(1))])
            if part := line[j:]:
                e.children.append(self._e(part))

        return g

    def left_marker(self, pos: Vec):
        w = self.settings.marker_width
        h = self.settings.marker_projected_height
        dh = 2 * self.settings.marker_projected_height

        path = self._elem.elem("path")

        if self.settings.end_class == EndClass.SIMPLE:
            path.attrs["d"] = f"M{pos.x} {pos.y}h{w}m{-dh} {-h}v{dh}"
        elif self.settings.end_class == EndClass.COMPLEX:
            path.attrs["d"] = f"M{pos.x} {pos.y}h{w}m{-dh} {-h}v{dh}m10 {-dh}v{dh}"
        else:
            raise NotImplementedError(f"unknown end class {self.settings.end_class}")

    def right_marker(self, pos: Vec):
        w = self.settings.marker_width
        h = self.settings.marker_projected_height
        dh = 2 * self.settings.marker_projected_height

        path = self._elem.elem("path")

        if self.settings.end_class == EndClass.SIMPLE:
            path.attrs["d"] = f"M{pos.x} {pos.y}h{w}m0 {-h}v{dh}"
        elif self.settings.end_class == EndClass.COMPLEX:
            path.attrs["d"] = f"M{pos.x} {pos.y}h{w}m0 {-h}v{dh}m-10 {-dh}v{dh}"
        else:
            raise NotImplementedError(f"unknown end class {self.settings.end_class}")

    def debug(self, node: Element[_t.Any], context: RenderContext):
        if not self._debug:
            return

        if not context.reverse:
            pos = context.pos

            left_padding_box = pos.x
            right_padding_box = pos.x + node.width

            left_display_box = pos.x
            right_display_box = pos.x + node.display_width

            left_content_box = left_padding_box + node.start_padding
            right_content_box = right_padding_box - node.end_padding

            left_margin_box = left_content_box - node.start_margin
            right_margin_box = right_content_box + node.end_margin
        else:
            pos = context.pos - Vec(node.width, 0)

            left_padding_box = pos.x
            right_padding_box = pos.x + node.width

            left_display_box = pos.x
            right_display_box = pos.x + node.display_width

            left_content_box = left_padding_box + node.end_padding
            right_content_box = right_padding_box - node.start_padding

            left_margin_box = left_content_box - node.end_margin
            right_margin_box = right_content_box + node.start_margin

        self._debug_box(
            left_display_box,
            right_display_box,
            pos.y - node.up,
            node.up + node.height + node.down,
            "dbg-display",
        )
        self._debug_box(
            left_content_box,
            right_content_box,
            pos.y - node.up,
            node.up,
            "dbg-content",
        )
        self._debug_box(
            left_content_box, right_content_box, pos.y, node.height, "dbg-content-main"
        )
        self._debug_box(
            left_content_box,
            right_content_box,
            pos.y + node.height,
            node.down,
            "dbg-content",
        )
        self._debug_box(
            left_padding_box,
            right_padding_box,
            pos.y - node.up,
            node.up + node.height + node.down,
            "dbg-padding",
        )
        self._debug_box(
            left_margin_box,
            right_margin_box,
            pos.y - node.up,
            node.up + node.height + node.down,
            "dbg-margin",
        )

    def _debug_box(self, left: float, right: float, y: float, h: float, css_class: str):
        if h == 0:
            y -= 0.5
            h = 0.5
        self._elem.elem(
            "rect",
            {
                "x": left,
                "y": y,
                "width": right - left,
                "height": h,
                "class": css_class,
            },
        )

    def debug_pos(self, pos: Vec, css_class: str = ""):
        if not self._debug:
            return

        if css_class:
            css_class = f"dbg-position {css_class}"
        else:
            css_class = "dbg-position"
        self._elem.elem(
            "path",
            {
                "d": f"M{pos.x} {pos.y}h-5h10h-5v-5v10v-5",
                "class": css_class,
            },
        )

    def debug_ridge_line(self, pos: Vec, node: Element[_t.Any], reverse: bool):
        if not self._debug:
            return

        if not reverse:
            start = 0
            end = self._width
            dir = 1
        else:
            start = self._width
            end = 0
            dir = -1

        d = f"M{start} {pos.y - node.top_ridge_line.before}"
        for p in node.top_ridge_line.ridge:
            d += f"H{pos.x + dir * p.x}V{pos.y - p.y}"
        d += f"H{end}"
        pos = pos + Vec(0, node.height)
        for p in reversed(node.bottom_ridge_line.ridge):
            d += f"V{pos.y + p.y}H{pos.x + dir * p.x}"
        d += f"V{pos.y + node.bottom_ridge_line.before}H{start}Z"

        self._elem.elem(
            "path",
            {
                "d": d,
                "class": "dbg-ridge-line",
            },
        )

    _ESCAPE_RE = re.compile(r"[*_`\[\]<&\"]", re.UNICODE)

    def _e(self, text: _t.Any):
        return self._ESCAPE_RE.sub(lambda c: f"&#{ord(c[0])};", str(text))

    class _SvgLine(Line):
        def __init__(
            self, render: SvgRender[_t.Any], pos: Vec, reverse: bool, css_class: str
        ):
            self._render = render
            self._pos = Vec(pos.x, pos.y)
            self._reverse = reverse
            self._elem = render._elem.elem("path")
            self._elem.attrs["d"] = f"M{pos.x} {pos.y}"
            if css_class:
                self._elem.attrs["class"] = css_class

        def segment_abs(
            self, x: int, arrow_begin: bool = False, arrow_end: bool = False
        ) -> Line:
            w = x - self._pos.x

            if arrow_begin and abs(w) >= self._render._arrow_length:
                self._arrow("e" if w >= 0 else "w")

            self._elem.attrs["d"] += f"H{x}"
            self._pos.x = x

            if arrow_end and abs(w) >= self._render._arrow_length:
                self._arrow("e" if w >= 0 else "w", end=True)

            return self

        def _arrow(self, d: Direction, end: bool = False):
            if self._render._arrow_style is ArrowStyle.NONE:
                return
            transform = f"translate({self._pos.x}, {self._pos.y})"
            if d == "w":
                transform += f" scale(-1, 1)"
            if not end:
                transform += f" translate({self._render._arrow_length}, 0)"
            self._render._elem.elem(
                "use",
                {
                    "href": f"#{self._render._arrow_id}",
                    "class": self._render._arrow_class,
                    "transform": transform,
                },
            )

        def bend(
            self,
            y: int,
            coming_from: Direction,
            coming_to: Direction | None,
            arrow_begin: bool = False,
            arrow_end: bool = False,
        ):
            h = y - self._pos.y

            double_arc_radius = math.ceil(2 * self._render.settings.arc_radius)

            if abs(h) < double_arc_radius:
                return self._bend_bezier(h, coming_from, coming_to)
            elif h > 0:
                h -= double_arc_radius
                intermediate_d = ("s", "n")
            else:
                h += double_arc_radius
                intermediate_d = ("n", "s")

            self._arc(coming_from, intermediate_d[0])

            if coming_to is not None:
                self._elem.attrs["d"] += f"v{h}"
                self._pos.y += h
                self._arc(intermediate_d[1], coming_to)
            else:
                arc_radius = math.ceil(self._render.settings.arc_radius)
                self._elem.attrs["d"] += f"v{h + arc_radius}"
                self._pos.y += h + arc_radius

            return self

        def _arc(self, coming_from: str, coming_to: str):
            arc_radius = math.ceil(self._render.settings.arc_radius)

            x = arc_radius
            y = arc_radius
            if coming_from == "e" or coming_to == "w":
                x = -x
            if coming_from == "s" or coming_to == "n":
                y = -y

            sf = (
                0
                if (coming_from, coming_to)
                in [("n", "e"), ("e", "s"), ("s", "w"), ("w", "n")]
                else 1
            )

            self._elem.attrs["d"] += f"a{arc_radius} {arc_radius} 0 0 {sf} {x} {y}"

            self._pos.x += x
            self._pos.y += y

            return self

        def _bend_bezier(self, h: int, coming_from: str, coming_to: str | None):
            double_arc_radius = math.ceil(2 * self._render.settings.arc_radius)

            interm_x_1: float = self._pos.x
            interm_y_1: float = self._pos.y
            interm_x_2: float = self._pos.x
            interm_y_2: float = self._pos.y + h
            out_x = self._pos.x
            out_y = self._pos.y + h

            if coming_from == "w" and coming_to == "e":
                interm_x_1 += 2 * double_arc_radius / 3
                interm_x_2 += double_arc_radius / 3
                out_x += double_arc_radius
            elif coming_from == "e" and coming_to == "w":
                interm_x_1 -= 2 * double_arc_radius / 3
                interm_x_2 -= double_arc_radius / 3
                out_x -= double_arc_radius
            elif coming_from == coming_to == "w":
                interm_x_1 += 2 * double_arc_radius / 3
                interm_x_2 += 2 * double_arc_radius / 3
            elif coming_from == coming_to == "e":
                interm_x_1 -= 2 * double_arc_radius / 3
                interm_x_2 -= 2 * double_arc_radius / 3
            elif coming_from == "w":
                interm_x_1 += double_arc_radius / 2
                interm_x_2 += double_arc_radius / 2
                interm_y_2 = self._pos.y + h / 2
                out_x += math.ceil(self._render.settings.arc_radius)
            elif coming_from == "e":
                interm_x_1 -= double_arc_radius / 2
                interm_x_2 -= double_arc_radius / 2
                interm_y_2 = self._pos.y + h / 2
                out_x -= math.ceil(self._render.settings.arc_radius)

            self._elem.attrs[
                "d"
            ] += f"C{interm_x_1} {interm_y_1} {interm_x_2} {interm_y_2} {out_x} {out_y}"

            self._pos.x = out_x
            self._pos.y += h

            return self

    @dataclass
    class _SvgElement:
        name: str
        """Name of SVG node"""

        attrs: dict[str, _t.Any] = field(default_factory=dict)
        """SVG node attributes"""

        children: list[SvgRender._SvgElement | str] = field(default_factory=list)
        """Children SVG nodes"""

        def elem(
            self,
            name: str,
            attrs: dict[str, _t.Any] | None = None,
            children: list[SvgRender._SvgElement | str] | None = None,
        ):
            return SvgRender._SvgElement(name, attrs or {}, children or []).add_to(self)

        def add_to(self, parent: SvgRender._SvgElement) -> SvgRender._SvgElement:
            parent.children.append(self)
            return self

        def write_svg(self, f: _t.TextIO, render: SvgRender[_t.Any]):
            f.write(f"<{self.name}")
            for name, value in sorted(self.attrs.items()):
                if value is not None:
                    f.write(f' {name}="{render._e(value)}"')
            f.write(">")
            for child in self.children:
                if isinstance(child, SvgRender._SvgElement):
                    child.write_svg(f, render)
                else:
                    f.write(str(child))
            f.write(f"</{self.name}>")



from __future__ import annotations

from dataclasses import dataclass


@dataclass(slots=True)
class Vec:
    x: int
    y: int

    def __add__(self, rhs: Vec):
        return Vec(self.x + rhs.x, self.y + rhs.y)

    def __sub__(self, rhs: Vec):
        return Vec(self.x - rhs.x, self.y - rhs.y)

    def __mul__(self, rhs: int):
        return Vec(self.x * rhs, self.y * rhs)



from __future__ import annotations

import abc
import math
import pathlib
import typing as _t

import grapheme
import wcwidth

try:
    from PIL.ImageFont import FreeTypeFont
except ImportError:
    pass

__all__ = [
    "TextMeasure",
    "SimpleTextMeasure",
    "TrueTextMeasure",
]


class TextMeasure(metaclass=abc.ABCMeta):
    """
    An interface for measuring dimensions of rendered text.

    """

    @abc.abstractmethod
    def measure(self, text: str) -> _t.Tuple[int, int]:
        """
        Called before rendering a node to measure its text.

        Should return a tuple ``(width, height)``.

        """

    @property
    @abc.abstractmethod
    def font_size(self) -> float:
        """
        Font size.

        """

    @property
    @abc.abstractmethod
    def line_height(self) -> float:
        """
        Height of a single line, equals to CSS ``line-height`` property.

        """

    @property
    @abc.abstractmethod
    def ascent(self) -> float:
        """
        Distance from the baseline to the highest outline point.

        """


class SimpleTextMeasure(TextMeasure):
    """
    A simple text measuring service that multiplies length of the text
    by average width of the character. Works best for monospace fonts.

    """

    def __init__(
        self,
        *,
        character_advance: float,
        wide_character_advance: float,
        font_size: float,
        line_height: float,
        ascent: float,
    ):
        """
        :param character_advance:
            average length of one character in the used font. Since SVG elements
            cannot expand and shrink dynamically, length of text nodes is calculated
            as number of symbols multiplied by this constant.

        :param wide_character_advance:
            average length of one wide character in the used font.

        :param font_size:
            font size.

        :param line_height:
            height of a single line, equals to CSS ``line-height`` property.

        :param ascent:
            the distance from the baseline to the highest outline point.

            This is required because Chrome, Firefox and Safari all treat
            ``dominant-baseline`` attribute in a slightly different way.

        """

        self._character_advance = character_advance
        self._wide_character_advance = wide_character_advance
        self._font_size = font_size
        self._line_height = line_height
        self._ascent = ascent

    def measure(self, text: str) -> _t.Tuple[int, int]:
        if not text:
            return (0, math.ceil(self._line_height))

        lines = text.splitlines()
        line_width = math.ceil(
            max(
                sum(
                    (
                        self._character_advance
                        if (width := wcwidth.wcswidth(g)) == 1
                        else (self._wide_character_advance if width == 2 else 0)
                    )
                    for g in grapheme.graphemes(line)
                )
                for line in lines
            )
        )

        return (line_width, math.ceil(len(lines) * self._line_height))

    @property
    def font_size(self) -> float:
        return self._font_size

    @property
    def line_height(self) -> float:
        return self._line_height

    @property
    def ascent(self) -> float:
        return self._ascent

    def __eq__(self, rhs):
        if not isinstance(rhs, SimpleTextMeasure):
            return False
        else:
            return (
                self._character_advance,
                self._wide_character_advance,
                self._font_size,
                self._line_height,
            ) == (
                rhs._character_advance,
                rhs._wide_character_advance,
                rhs._font_size,
                rhs._line_height,
            )


class TrueTextMeasure(TextMeasure):
    """
    A text measuring service that uses Python Image Library to calculate
    text's actual size.

    """

    def __init__(
        self,
        font: FreeTypeFont | str | pathlib.Path,
        font_size: float,
        line_height: float,
    ):
        """
        :param font:
            font data, a path to a font file, or a font file name.

            .. seealso::

                `PIL.ImageFont.ImageFont`, `PIL.ImageFont.truetype`

        :param font_size:
            font size.

        :param line_height:
            height of a single line, equals to CSS ``line-height`` property.

        """

        self._font = font

        if isinstance(font, (str, pathlib.Path)):
            import PIL.ImageFont

            self._font_data = PIL.ImageFont.truetype(font, size=font_size)
        else:
            self._font_data = font

        self._font_size = font_size
        self._line_height = line_height

        self._ascent = self._font_data.getmetrics()[0]

    def measure(self, text: str) -> _t.Tuple[int, int]:
        if not text:
            return (0, math.ceil(self._line_height))

        lines = text.splitlines()
        line_width = math.ceil(max(self._font_data.getlength(line) for line in lines))
        return (line_width, math.ceil(len(lines) * self._line_height))

    @property
    def font_size(self) -> float:
        return self._font_size

    @property
    def ascent(self) -> float:
        return self._ascent

    @property
    def line_height(self) -> float:
        return self._line_height

    def __eq__(self, rhs):
        if not isinstance(rhs, TrueTextMeasure):
            return False
        else:
            return (
                self._font,
                self._font_size,
                self._line_height,
            ) == (
                rhs._font,
                rhs._font_size,
                rhs._line_height,
            )


def _main():
    import argparse
    import textwrap

    import PIL.ImageFont

    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--size", type=int, default=14)
    parser.add_argument("font", type=argparse.FileType("rb"))
    args = parser.parse_args()

    tm = TrueTextMeasure(
        PIL.ImageFont.truetype(args.font, size=args.size), args.size, 1.1 * args.size
    )
    character_advance = tm.measure("x" * 1000)[0] / 1000
    wide_character_advance = tm.measure("🥲" * 1000)[0] / 1000
    font_size = args.size
    line_height = 1.1 * args.size
    ascent = tm.ascent

    print(
        textwrap.dedent(
            f"""
                SimpleTextMeasure(
                    character_advance={character_advance!r},
                    wide_character_advance={wide_character_advance!r},
                    font_size={font_size!r},
                    line_height={line_height!r},
                    ascent={ascent!r},
                )
            """
        ).strip()
    )


if __name__ == "__main__":
    _main()



from __future__ import annotations

import itertools
import math
import typing as _t
from dataclasses import replace
from functools import cached_property

from syntax_diagrams._impl.render import (
    ConnectionDirection,
    ConnectionType,
    LayoutContext,
    LayoutSettings,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.ridge_line import RidgeLine, merge_ridge_lines
from syntax_diagrams._impl.tree import Element
from syntax_diagrams._impl.tree.choice import Choice
from syntax_diagrams._impl.tree.skip import Skip
from syntax_diagrams._impl.vec import Vec
from syntax_diagrams.element import LineBreak

T = _t.TypeVar("T")


class Sequence(Element[T], _t.Generic[T]):
    _items: list[Element[T]]
    _linebreaks: list[LineBreak]

    _item_rows: list[list[Element[T]]]
    _item_rows_layout: list[tuple[int | None, int, int]]
    _start_connection: ConnectionType
    _end_connection: ConnectionType
    _shift_first_line: bool

    def __new__(
        cls,
        items: list[Element[T]],
        linebreaks: list[LineBreak] | LineBreak | None = None,
    ) -> Element[T]:
        if len(items) == 1:
            return items[0]
        elif len(items) == 0 or all(isinstance(item, Skip) for item in items):
            return Skip()

        if isinstance(linebreaks, LineBreak):
            linebreaks = [linebreaks] * (len(items) - 1)
        elif linebreaks is None:
            linebreaks = [LineBreak.DEFAULT] * (len(items) - 1)
        if len(items) != len(linebreaks) + 1:
            raise ValueError(
                f"number of line breaks (={len(linebreaks)}) "
                f"must be one less than number of items (={len(items)})"
            )

        self = super().__new__(cls)

        self._items = []
        self._linebreaks = []

        for item, linebreak in itertools.zip_longest(items, linebreaks):
            if isinstance(item, Skip):
                continue
            if isinstance(item, Sequence):
                self._items.extend(item._items)  # type: ignore
                self._linebreaks.extend(item._linebreaks)
            else:
                self._items.append(item)
            if linebreak is not None:
                self._linebreaks.append(linebreak)

        return self

    @cached_property
    def precedence(self) -> int:
        return 2

    @cached_property
    def contains_choices(self) -> bool:
        return any(item.contains_choices for item in self._items)

    @cached_property
    def can_use_opt_enters(self) -> bool:
        return self._items[0].can_use_opt_enters

    @cached_property
    def can_use_opt_exits(self) -> bool:
        return self._items[-1].can_use_opt_exits

    def _calculate_content_layout(
        self, settings: LayoutSettings[T], context: LayoutContext
    ):
        if LineBreak.NO_BREAK in self._linebreaks and not all(
            linebreak is LineBreak.NO_BREAK for linebreak in self._linebreaks
        ):
            self._join_no_breaks()

        if LineBreak.HARD not in self._linebreaks:
            single_line_width, gaps_width = self._calculate_layout_single_line(
                settings, context
            )
            if single_line_width <= context.width:
                self._item_rows = [self._items]
                self._calculate_layout_metrics(settings, context)
                return
            elif all(linebreak is LineBreak.NO_BREAK for linebreak in self._linebreaks):
                if gaps_width < context.width and context.width > 0:
                    scale = (context.width - gaps_width) / (
                        single_line_width - gaps_width
                    )
                    self._calculate_layout_single_line(settings, context, scale)
                self._item_rows = [self._items]
                self._calculate_layout_metrics(settings, context)
                return

        # Isolate stack to avoid intersections. We could be smarter and avoid isolation
        # by adjusting starting positions of continuation lines. However,
        # the implementation is complicated as it is now, so I'd rather not do this.
        context = self._isolate(
            context.start_direction is ConnectionDirection.DOWN
            or not context.start_bottom_is_clear,
            context.end_direction is ConnectionDirection.UP
            or not context.end_top_is_clear,
        )
        self._calculate_layout_multi_line(settings, context)
        self._calculate_layout_metrics(settings, context)

    def _join_no_breaks(self):
        new_items: list[Element[T]] = []
        new_linebreaks: list[LineBreak] = []
        current_seq: list[Element[T]] = []

        for item, linebreak in itertools.zip_longest(self._items, self._linebreaks):
            current_seq.append(item)
            if linebreak is not LineBreak.NO_BREAK:
                new_items.append(Sequence(current_seq, LineBreak.NO_BREAK))
                new_linebreaks.append(linebreak)
                current_seq.clear()

        self._items = new_items
        self._linebreaks = new_linebreaks

    def _calculate_layout_single_line(
        self,
        settings: LayoutSettings[T],
        context: LayoutContext,
        scale: float | None = None,
    ):
        self._start_connection = context.start_connection
        self._end_connection = context.end_connection
        self._shift_first_line = False

        width = 0
        gaps_width = 0

        for i, item in enumerate(self._items):
            if scale is not None:
                max_width = math.floor(item.display_width * scale)
            else:
                max_width = context.width
            item_context = LayoutContext(
                width=max_width,
                is_outer=context.is_outer,
                start_top_is_clear=True,
                start_bottom_is_clear=True,
                start_direction=ConnectionDirection.STRAIGHT,
                end_top_is_clear=True,
                end_bottom_is_clear=True,
                end_direction=ConnectionDirection.STRAIGHT,
                allow_shrinking_stacks=False,
            )
            if i == 0:
                item_context.opt_enter_top = context.opt_enter_top
                item_context.opt_enter_bottom = context.opt_enter_bottom
                item_context.start_connection = self._start_connection
                item_context.start_top_is_clear = context.start_top_is_clear
                item_context.start_bottom_is_clear = context.start_bottom_is_clear
            if i == len(self._items) - 1:
                item_context.opt_exit_top = context.opt_exit_top
                item_context.opt_exit_bottom = context.opt_exit_bottom
                item_context.end_connection = self._end_connection
                item_context.end_top_is_clear = context.end_top_is_clear
                item_context.end_bottom_is_clear = context.end_bottom_is_clear
                item_context.allow_shrinking_stacks = context.allow_shrinking_stacks
            item.calculate_layout(settings, item_context)
            if i > 0:
                gap = self._calculate_gap(self._items[i - 1], item, settings)
                width += gap
                gaps_width += gap
            width += item.display_width
        return width, gaps_width

    def _calculate_layout_multi_line(
        self,
        settings: LayoutSettings[T],
        context: LayoutContext,
    ):
        arc_size = ConnectionType.STACK.arc_size(settings)

        # Like `NORMAL`, but adds extra gap to line up nodes in the first
        # and subsequent rows. This nicely lines up all first nodes in all rows.
        self._start_connection = (
            ConnectionType.STACK_BOUND
            if context.start_connection is ConnectionType.NORMAL
            else context.start_connection
        )
        # `context.end_connection` is `NORMAL` because we're isolated.
        self._end_connection = context.end_connection
        # Shift all lines except first by one arc radius to line them up
        # with the first one.
        self._shift_first_line = (
            isinstance(self._items[0], Choice)
            or self._start_connection is ConnectionType.SPLIT
        )

        self._item_rows = []

        current_row: list[Element[T]] = []
        current_width = 0

        last_soft_break_idx = 0
        width_at_last_soft_break = 0
        margin_after_last_soft_break = 0

        # Adjust line width for first line shift.
        if self._shift_first_line:
            max_width = context.width - math.floor(settings.arc_radius)
        else:
            max_width = context.width
        max_line_width = context.width

        for i, (item, linebreak) in enumerate(
            itertools.zip_longest(self._items, self._linebreaks)
        ):

            # Prepare item's layout context.
            item_context = LayoutContext(
                width=max_line_width,
                is_outer=context.is_outer,
                start_top_is_clear=True,
                start_bottom_is_clear=True,
                start_direction=ConnectionDirection.STRAIGHT,
                end_top_is_clear=True,
                end_bottom_is_clear=True,
                end_direction=ConnectionDirection.STRAIGHT,
                allow_shrinking_stacks=False,
            )
            if i == 0:
                item_context.opt_enter_top = context.opt_enter_top
                item_context.start_connection = self._start_connection
                item_context.start_top_is_clear = context.start_top_is_clear
                item_context.start_bottom_is_clear = context.start_bottom_is_clear
                item_context.start_direction = context.start_direction
            elif not current_row:
                item_context.start_connection = ConnectionType.STACK
                item_context.start_direction = ConnectionDirection.UP
            if i == len(self._items) - 1:
                item_context.opt_exit_bottom = context.opt_exit_bottom
                item_context.end_connection = self._end_connection
                item_context.end_top_is_clear = context.end_top_is_clear
                item_context.end_bottom_is_clear = context.end_bottom_is_clear
                item_context.end_direction = context.end_direction
                item_context.allow_shrinking_stacks = context.allow_shrinking_stacks

            # Calculate item's layout.
            item.calculate_layout(settings, item_context)

            # Calculate margin between current node and the previous one, if any.
            if current_row:
                margin = self._calculate_gap(current_row[-1], item, settings)
            else:
                margin = 0

            # Make a soft break, if needed and possible.
            if (
                current_row
                and current_width + margin + item.display_width + arc_size
                > max_line_width
                and last_soft_break_idx
            ):
                self._item_rows.append(current_row[:last_soft_break_idx])
                current_row = current_row[last_soft_break_idx:]
                current_width -= width_at_last_soft_break + margin_after_last_soft_break
                last_soft_break_idx = 0
                width_at_last_soft_break = 0
                margin_after_last_soft_break = 0
                max_line_width = max_width
                item_context = replace(
                    item_context,
                    width=max_line_width,
                    start_connection=ConnectionType.STACK,
                    start_direction=ConnectionDirection.UP,
                )
                if current_row:
                    # Item at `last_soft_break_idx` is now first in its row.
                    # Recalculate its layout to account for stack connection.
                    current_width -= current_row[0].display_width
                    current_row[0].calculate_layout(settings, item_context)
                    current_width += current_row[0].display_width
                    margin = self._calculate_gap(current_row[-1], item, settings)
                else:
                    # Current item is now first in its row. Recalculate its layout
                    # to account for stack connection.
                    item.calculate_layout(settings, item_context)
                    margin = 0

            # If we still need a break, then a soft break wasn't possible
            # (or enough), so we'll have to break here.
            if (
                current_row
                and current_width + margin + item.display_width + arc_size
                > max_line_width
            ):
                self._item_rows.append(current_row)

                current_row = []
                current_width = 0
                last_soft_break_idx = 0
                width_at_last_soft_break = 0
                margin_after_last_soft_break = 0
                margin = 0
                max_line_width = max_width
                item_context = replace(
                    item_context,
                    width=max_line_width,
                    start_connection=ConnectionType.STACK,
                    start_direction=ConnectionDirection.UP,
                )
                # Current item is now first in its row. Recalculate its layout
                # to account for stack connection
                item.calculate_layout(settings, item_context)

            current_row.append(item)
            current_width += margin + item.display_width

            if linebreak is LineBreak.HARD:
                self._item_rows.append(current_row)
                current_row = []
                current_width = 0
                last_soft_break_idx = 0
                width_at_last_soft_break = 0
                margin_after_last_soft_break = 0
                max_line_width = max_width
            elif linebreak is LineBreak.SOFT:
                last_soft_break_idx = len(current_row)
                width_at_last_soft_break = current_width
            elif i == last_soft_break_idx:
                margin_after_last_soft_break = margin

        if current_row:
            self._item_rows.append(current_row)

        # Re-calculate layout for every last item of every row except for the last one.
        # This enables stack connections for them, and gives them an opportunity
        # to use carry line as optional exit.
        #
        # Consider `A B? C`:
        #
        #     ── A ─┬─ B ─╮
        #           ↓     ↓
        #     ╭─────┴─────╯  < note: carry line used as optional exit.
        #     ↓
        #     ╰─ C ──
        for i, row in enumerate(self._item_rows[:-1]):
            item_context = row[-1].context
            assert item_context
            item_context = replace(
                item_context,
                end_connection=ConnectionType.STACK,
                end_direction=ConnectionDirection.DOWN,
                opt_exit_bottom=True,
                allow_shrinking_stacks=True,
            )
            row[-1].calculate_layout(settings, item_context)

    def _calculate_layout_metrics(
        self, settings: LayoutSettings[T], context: LayoutContext
    ):
        self._item_rows_layout = []

        if context.is_outer:
            self._vertical_seq_separation = settings.vertical_seq_separation_outer
        else:
            self._vertical_seq_separation = settings.vertical_seq_separation

        self._line_shift = (
            math.floor(settings.arc_radius) if self._shift_first_line else 0
        )

        # Total element's width.
        width = 0
        display_width = 0
        # Width of the last row.
        row_width = 0
        # Padding between stack's left border and nearest content box.
        start_padding = None
        # Leftmost margin relative to the stack's left border. Can be negative.
        start_margin_offset = None
        # Rightmost padding relative to the stack's left border.
        end_padding_offset = None
        # Rightmost margin relative to the stack's left border.
        end_margin_offset = None

        pos = 0

        for i, row in enumerate(self._item_rows):
            row_up = 0
            row_down_offset = 0
            row_width = 0
            row_pos = 0

            line_shift = self._line_shift if i > 0 else 0
            row_width += line_shift

            for j, item in enumerate(row):
                if j == 0:
                    start_padding = (
                        item.start_padding + line_shift
                        if start_padding is None
                        else min(start_padding, item.start_padding + line_shift)
                    )
                    start_margin_offset = (
                        line_shift - item.start_margin + item.start_padding
                        if start_margin_offset is None
                        else min(
                            start_margin_offset,
                            line_shift - item.start_margin + item.start_padding,
                        )
                    )

                if j > 0:
                    gap = self._calculate_gap(row[j - 1], item, settings)
                    row_width += gap

                row_width += item.width

                if j == len(row) - 1:
                    end_padding_offset = (
                        row_width - item.end_padding
                        if end_padding_offset is None
                        else max(end_padding_offset, row_width - item.end_padding)
                    )
                    end_margin_offset = (
                        row_width + item.end_margin - item.end_padding
                        if end_margin_offset is None
                        else max(
                            end_margin_offset,
                            row_width + item.end_margin - item.end_padding,
                        )
                    )

                row_up = max(row_up, item.up - row_pos)
                row_pos += item.height
                row_down_offset = max(row_down_offset, row_pos + item.down)

            width = max(width, row_width)
            row_display_width = row_width - row[-1].width + row[-1].display_width
            display_width = max(display_width, row_display_width)

            if i > 0:
                row_upper_line = pos
                pos += self._vertical_seq_separation + row_up
                self._item_rows_layout.append((row_upper_line, pos, row_display_width))
            else:
                self.up = row_up
                self._item_rows_layout.append((None, pos, row_display_width))

            if i < len(self._item_rows) - 1:
                pos += row_down_offset + self._vertical_seq_separation
            else:
                pos += row_pos
                self.down = max(0, row_down_offset - row_pos)

        self.height = pos

        self.display_width = display_width

        start_padding = start_padding or 0
        if context.allow_shrinking_stacks:
            # We can use width of the last row instead of the width of the longest row.
            last_elem = self._item_rows[-1][-1]
            width = row_width
            end_padding = last_elem.end_padding
            end_margin_offset = width - last_elem.end_padding + last_elem.end_margin
        else:
            end_padding = max(0, width - (end_padding_offset or 0))

        self.content_width = max(0, width - start_padding - end_padding)
        self.start_padding = start_padding
        self.end_padding = end_padding
        if start_margin_offset is None:
            self.start_margin = 0
        else:
            self.start_margin = max(0, (-start_margin_offset) + start_padding)
        if end_margin_offset is None:
            self.end_margin = 0
        else:
            self.end_margin = max(0, (end_margin_offset - width) + end_padding)

    def _render_content(self, render: Render[T], context: RenderContext):
        arc_radius = math.ceil(render.settings.arc_radius)

        for i, row in enumerate(self._item_rows):
            row_upper_line, row_pos, _ = self._item_rows_layout[i]
            row_lower_line, next_row_pos, _ = (
                self._item_rows_layout[i + 1]
                if i < len(self._item_rows) - 1
                else (None, None, None)
            )

            if i == 0:
                pos = context.pos + Vec(0, row_pos)
            else:
                pos = context.pos + Vec(context.dir * self._line_shift, row_pos)

            for j, item in enumerate(row):
                if j > 0:
                    gap = self._calculate_gap(row[j - 1], item, render.settings)
                    render.line(pos, context.reverse).segment_abs(
                        pos.x + context.dir * gap
                    )
                    pos = pos + Vec(context.dir * gap, 0)

                if j == 0 and row_upper_line:
                    start_connection_pos = context.pos + Vec(
                        context.dir * (self._line_shift + arc_radius),
                        row_upper_line,
                    )
                elif j == 0:
                    start_connection_pos = context.start_connection_pos
                else:
                    start_connection_pos = pos

                end_pos = pos + Vec(context.dir * item.width, item.height)

                if j == len(row) - 1 and row_lower_line:
                    end_connection_pos = Vec(
                        end_pos.x - context.dir * arc_radius,
                        context.pos.y + row_lower_line,
                    )
                elif j == len(row) - 1:
                    end_connection_pos = context.end_connection_pos
                else:
                    end_connection_pos = end_pos

                item_context = RenderContext(
                    pos=pos,
                    start_connection_pos=start_connection_pos,
                    end_connection_pos=end_connection_pos,
                    reverse=context.reverse,
                )

                if i == 0 and j == 0:
                    if row_upper_line is None:
                        item_context.opt_enter_top = context.opt_enter_top
                    if row_lower_line is None:
                        item_context.opt_enter_bottom = context.opt_enter_bottom
                if i == len(self._item_rows) - 1 and j == len(row) - 1:
                    if row_upper_line is None:
                        item_context.opt_exit_top = context.opt_exit_top
                    if row_lower_line is None:
                        item_context.opt_exit_bottom = context.opt_exit_bottom
                if j == len(row) - 1 and row_lower_line is not None:
                    assert next_row_pos is not None
                    item_context.opt_exit_bottom = (
                        "w" if not context.reverse else "e",
                        (
                            context.pos
                            + Vec(
                                context.dir * (self._line_shift + arc_radius),
                                row_lower_line,
                            )
                        ),
                        (
                            context.pos
                            + Vec(
                                context.dir * self._line_shift,
                                row_lower_line
                                + min(
                                    arc_radius,
                                    math.ceil((next_row_pos - row_lower_line) / 2),
                                ),
                            )
                        ),
                    )

                item.render(render, item_context)

                pos = end_pos

            if row_lower_line is not None:
                line_start = Vec(
                    pos.x - context.dir * arc_radius,
                    context.pos.y + row_lower_line,
                )
                (
                    (render)
                    .line(line_start, context.reverse)
                    .segment_abs(
                        context.dir * self._line_shift
                        + context.pos.x
                        + context.dir * arc_radius,
                        arrow_begin=True,
                        arrow_end=True,
                    )
                )

    def _calculate_top_ridge_line(self) -> RidgeLine:
        assert self.settings

        seen_width = 0

        result = RidgeLine(-self.height, [])

        for i, row in enumerate(self._item_rows):
            _, row_pos, row_display_width = self._item_rows_layout[i]

            if row_display_width < seen_width:
                continue

            if i == 0:
                pos = Vec(0, -row_pos)
            else:
                pos = Vec(self._line_shift, -row_pos)

            row_ridge = RidgeLine(-self.height, [])

            for j, item in enumerate(row):
                if j > 0:
                    gap = self._calculate_gap(row[j - 1], item, self.settings)
                    pos = pos + Vec(gap, 0)

                item_ridge = item.top_ridge_line + pos
                row_ridge = merge_ridge_lines(row_ridge, item_ridge)

                end_pos = pos + Vec(item.width, -item.height)

                pos = end_pos

            seen_width = max(seen_width, row_display_width)

            row_ridge = merge_ridge_lines(
                row_ridge,
                RidgeLine(
                    self.up,
                    [
                        Vec(row_display_width, -self.height),
                    ],
                ),
                cmp=min,
            )
            result = merge_ridge_lines(result, row_ridge)

        return result

    def _calculate_bottom_ridge_line(self) -> RidgeLine:
        assert self.settings

        row = self._item_rows[-1]

        pos = Vec(0, self._item_rows_layout[-1][1] - self.height)
        before = pos.y
        if len(self._item_rows) > 1:
            pos = pos + Vec(self._line_shift, 0)

        result = RidgeLine(pos.y, [])
        for i, item in enumerate(row):
            if i > 0:
                gap = self._calculate_gap(row[i - 1], item, self.settings)
                pos = pos + Vec(gap, 0)
            pos = pos + Vec(0, item.height)
            item_ridge = item.bottom_ridge_line + pos
            item_ridge.before = before
            result = merge_ridge_lines(result, item_ridge)
            pos = pos + Vec(item.width, 0)

        return merge_ridge_lines(
            result,
            RidgeLine(
                -self.height,
                [
                    Vec(0, self.down),
                    Vec(self.display_width, 0),
                ],
            ),
            cmp=min,
        )

    @staticmethod
    def _calculate_gap(
        prev: Element[T], next: Element[T], settings: LayoutSettings[_t.Any]
    ) -> int:
        prev_gap = prev.end_margin - prev.end_padding - next.start_padding
        next_gap = next.start_margin - next.start_padding - prev.end_padding
        return max(0, prev_gap, next_gap, settings.arc_margin)

    def __str__(self):
        return " ".join(
            f"{item}" if item.precedence >= self.precedence else f"({item})"
            for item in self._items
        )



from __future__ import annotations

import typing as _t
from types import NoneType

from syntax_diagrams._impl.render import NodeStyle
from syntax_diagrams._impl.tree import Element as _Output
from syntax_diagrams._impl.tree.barrier import Barrier
from syntax_diagrams._impl.tree.choice import Choice
from syntax_diagrams._impl.tree.group import Group
from syntax_diagrams._impl.tree.node import Node
from syntax_diagrams._impl.tree.one_or_more import OneOrMore
from syntax_diagrams._impl.tree.sequence import Sequence
from syntax_diagrams._impl.tree.skip import Skip
from syntax_diagrams.element import Element as _Input
from syntax_diagrams.element import LineBreak
from syntax_diagrams.render import LoadingError

T = _t.TypeVar("T")


def load(
    element: _Input[T], convert_resolver_data: _t.Callable[[_t.Any | None], T | None]
) -> _Output[T]:
    if element is None:
        return Skip()
    elif isinstance(element, str):
        return _load_terminal(element, {}, convert_resolver_data)
    elif isinstance(element, list):
        return _load_sequence(element, {}, convert_resolver_data)
    elif isinstance(element, dict):
        ctors = {
            "sequence": _load_sequence,
            "stack": _load_stack,
            "no_break": _load_no_break,
            "choice": _load_choice,
            "optional": _load_optional,
            "one_or_more": _load_one_or_more,
            "zero_or_more": _load_zero_or_more,
            "barrier": _load_barrier,
            "terminal": _load_terminal,
            "non_terminal": _load_non_terminal,
            "comment": _load_comment,
            "group": _load_group,
        }

        ctors_found: list[str] = []

        for name in element:
            if name in ctors:
                ctors_found.append(name)

        if len(ctors_found) != 1:
            raise LoadingError(f"cannot determine type for {element!r}")

        name = ctors_found[0]
        element = element.copy()
        arg = element.pop(name)
        return ctors[name](
            arg, _t.cast(dict[str, _t.Any], element), convert_resolver_data
        )
    else:
        raise LoadingError(
            f"diagram item description should be string, "
            f"list or object, got {type(element)} instead"
        )


def _load_linebreaks(linebreak: _t.Any) -> LineBreak | list[LineBreak] | None:
    if linebreak is None:
        return None
    if isinstance(linebreak, LineBreak):
        return linebreak
    if isinstance(linebreak, str):
        return LineBreak(linebreak.upper())
    else:
        for i, x in enumerate(linebreak):
            _ensure_type(f"linebreaks[{i}]", x, str, LineBreak)
        return [LineBreak(x) for x in linebreak]


def _load_sequence(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = Sequence(
        [load(x, convert_resolver_data) for x in _ensure_type("sequence", a, list)],
        _load_linebreaks(kw.pop("linebreaks", None)),
    )
    _ensure_empty_dict("sequence", kw)
    return element


def _load_stack(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = Sequence(
        [load(x, convert_resolver_data) for x in _ensure_type("sequence", a, list)],
        LineBreak.HARD,
    )
    _ensure_empty_dict("sequence", kw)
    return element


def _load_no_break(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = Sequence(
        [load(x, convert_resolver_data) for x in _ensure_type("sequence", a, list)],
        LineBreak.NO_BREAK,
    )
    _ensure_empty_dict("sequence", kw)
    return element


def _load_choice(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = Choice(
        [load(x, convert_resolver_data) for x in _ensure_type("choice", a, list)],
        default=_ensure_type("default", kw.pop("default", None), int, NoneType) or 0,
    )
    _ensure_empty_dict("choice", kw)
    return element


def _load_optional(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = load(a, convert_resolver_data)
    skip = _ensure_type("skip", kw.pop("skip", None), bool, NoneType) or False
    skip_bottom = (
        _ensure_type("skip_bottom", kw.pop("skip_bottom", None), bool, NoneType)
        or False
    )
    _ensure_empty_dict("optional", kw)
    if skip_bottom:
        return Choice([element, Skip()], 1 if skip else 0)
    else:
        return Choice([Skip(), element], 0 if skip else 1)


def _load_one_or_more(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = OneOrMore(
        load(a, convert_resolver_data),
        repeat=load(kw.pop("repeat", None), convert_resolver_data),
        repeat_top=_ensure_type(
            "repeat_top", kw.pop("repeat_top", None), bool, NoneType
        )
        or False,
    )
    _ensure_empty_dict("one_or_more", kw)
    return element


def _load_zero_or_more(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = OneOrMore(
        load(a, convert_resolver_data),
        repeat=load(kw.pop("repeat", None), convert_resolver_data),
        repeat_top=_ensure_type(
            "repeat_top", kw.pop("repeat_top", None), bool, NoneType
        )
        or False,
    )
    skip = _ensure_type("skip", kw.pop("skip", None), bool, NoneType) or False
    skip_bottom = (
        _ensure_type("skip_bottom", kw.pop("skip_bottom", None), bool, NoneType)
        or False
    )
    _ensure_empty_dict("zero_or_more", kw)
    if skip_bottom:
        return Choice([element, Skip()], 1 if skip else 0)
    else:
        return Choice([Skip(), element], 0 if skip else 1)


def _load_barrier(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = Barrier(
        load(a, convert_resolver_data),
    )
    _ensure_empty_dict("barrier", kw)
    return element


def _load_terminal(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = Node(
        NodeStyle.TERMINAL,
        _ensure_type("terminal", a, str),
        href=_ensure_type("href", kw.pop("href", None), str, NoneType),
        title=_ensure_type("title", kw.pop("title", None), str, NoneType),
        css_class=_ensure_type("css_class", kw.pop("css_class", None), str, NoneType),
        resolve=_ensure_type("resolve", kw.pop("resolve", None), bool, NoneType),
        resolver_data=convert_resolver_data(kw.pop("resolver_data", None)),
    )
    _ensure_empty_dict("terminal", kw)
    return element


def _load_non_terminal(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = Node(
        NodeStyle.NON_TERMINAL,
        _ensure_type("non_terminal", a, str),
        href=_ensure_type("href", kw.pop("href", None), str, NoneType),
        title=_ensure_type("title", kw.pop("title", None), str, NoneType),
        css_class=_ensure_type("css_class", kw.pop("css_class", None), str, NoneType),
        resolve=_ensure_type("resolve", kw.pop("resolve", None), bool, NoneType),
        resolver_data=convert_resolver_data(kw.pop("resolver_data", None)),
    )
    _ensure_empty_dict("non_terminal", kw)
    return element


def _load_comment(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = Node(
        NodeStyle.COMMENT,
        _ensure_type("comment", a, str),
        href=_ensure_type("href", kw.pop("href", None), str, NoneType),
        title=_ensure_type("title", kw.pop("title", None), str, NoneType),
        css_class=_ensure_type("css_class", kw.pop("css_class", None), str, NoneType),
        resolve=_ensure_type("resolve", kw.pop("resolve", None), bool, NoneType),
        resolver_data=convert_resolver_data(kw.pop("resolver_data", None)),
    )
    _ensure_empty_dict("comment", kw)
    return element


def _load_group(
    a: _t.Any,
    kw: dict[str, _t.Any],
    convert_resolver_data: _t.Callable[[_t.Any | None], T | None],
) -> _Output[T]:
    element = Group(
        load(a, convert_resolver_data),
        text=_ensure_type("text", kw.pop("text", None), str, NoneType),
        href=_ensure_type("href", kw.pop("href", None), str, NoneType),
        title=_ensure_type("title", kw.pop("title", None), str, NoneType),
        css_class=_ensure_type("css_class", kw.pop("css_class", None), str, NoneType),
    )
    _ensure_empty_dict("comment", kw)
    return element


def _ensure_type(name: str, x: _t.Any, *types: type[T]) -> T:
    if not isinstance(x, types):
        types_str = ", ".join([t.__name__ for t in types])
        raise LoadingError(
            f"{name} should be {types_str}, got {type(x)} (={x!r}) instead"
        )
    return x


def _ensure_empty_dict(name: str, x: dict[str, _t.Any]):
    if x:
        keys = ", ".join(x.keys())
        raise LoadingError(f"{name} got unexpected parameters: {keys}")



from __future__ import annotations

import io
import itertools
import sys
import typing as _t

from syntax_diagrams._impl.render import (
    ConnectionDirection,
    ConnectionType,
    Direction,
    LayoutContext,
    LayoutSettings,
    Line,
    NodeStyle,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.tree import Element
from syntax_diagrams._impl.tree.barrier import Barrier
from syntax_diagrams._impl.tree.end import End
from syntax_diagrams._impl.tree.sequence import Sequence
from syntax_diagrams._impl.vec import Vec
from syntax_diagrams.element import LineBreak
from syntax_diagrams.measure import SimpleTextMeasure
from syntax_diagrams.render import EndClass, TextRenderSettings
from syntax_diagrams.resolver import HrefResolver

T = _t.TypeVar("T")


def render_text(
    node: Element[T],
    /,
    *,
    max_width: int | None = None,
    settings: TextRenderSettings = TextRenderSettings(),
    reverse: bool | None = None,
    href_resolver: HrefResolver[T] = HrefResolver(),
    dump_debug_data: bool = False,
) -> str:
    if max_width is None:
        max_width = settings.max_width
    if reverse is None:
        reverse = settings.reverse

    max_width = max(0, max_width - settings.padding[3] - settings.padding[1])

    node = Sequence(
        [End(), Barrier(Sequence([node, End(reverse=True)], LineBreak.NO_BREAK))],
        LineBreak.NO_BREAK,
    )

    layout_settings = text_layout_settings(settings)
    layout_settings.href_resolver = href_resolver
    node.calculate_layout(
        layout_settings,
        LayoutContext(
            width=max_width,
            is_outer=True,
            start_connection=ConnectionType.NULL,
            start_top_is_clear=True,
            start_bottom_is_clear=True,
            start_direction=ConnectionDirection.STRAIGHT,
            end_connection=ConnectionType.NULL,
            end_top_is_clear=True,
            end_bottom_is_clear=True,
            end_direction=ConnectionDirection.STRAIGHT,
            allow_shrinking_stacks=True,
        ),
    )

    render = TextRender(
        (settings.padding[3] + node.display_width + settings.padding[1]),
        (
            settings.padding[0]
            + node.up
            + node.height
            + node.down
            + settings.padding[2]
            + 1
        ),
        layout_settings,
        dump_debug_data,
    )

    pos = Vec(settings.padding[3], settings.padding[0] + node.up)
    if not reverse:
        start_connection_pos = pos
        end_connection_pos = pos + Vec(node.width, node.height)
    else:
        start_connection_pos = pos + Vec(node.display_width, 0)
        end_connection_pos = pos + Vec(node.display_width - node.width, node.height)

    node.render(
        render,
        RenderContext(
            pos=start_connection_pos,
            start_connection_pos=start_connection_pos,
            end_connection_pos=end_connection_pos,
            reverse=reverse,
        ),
    )

    if dump_debug_data:
        return render.debug_data()
    else:
        return render.to_string()


_TEXT_MEASURE = SimpleTextMeasure(
    character_advance=1,
    font_size=1,
    wide_character_advance=2,
    line_height=1,
    ascent=1,
)


def text_layout_settings(settings: TextRenderSettings = TextRenderSettings()):
    return LayoutSettings(
        horizontal_seq_separation=settings.horizontal_seq_separation,
        vertical_choice_separation_outer=settings.vertical_choice_separation_outer,
        vertical_choice_separation=settings.vertical_choice_separation,
        vertical_seq_separation_outer=settings.vertical_seq_separation_outer,
        vertical_seq_separation=settings.vertical_seq_separation,
        arc_radius=0.5,
        arc_margin=1,
        terminal_horizontal_padding=2,
        terminal_vertical_padding=0,
        terminal_radius=0,
        non_terminal_horizontal_padding=2,
        non_terminal_vertical_padding=0,
        non_terminal_radius=0,
        comment_horizontal_padding=2,
        comment_vertical_padding=0,
        comment_radius=0,
        terminal_text_measure=_TEXT_MEASURE,
        non_terminal_text_measure=_TEXT_MEASURE,
        comment_text_measure=_TEXT_MEASURE,
        group_text_measure=_TEXT_MEASURE,
        group_vertical_padding=settings.group_vertical_padding,
        group_horizontal_padding=settings.group_horizontal_padding,
        group_vertical_margin=settings.group_vertical_margin,
        group_horizontal_margin=settings.group_horizontal_margin,
        group_thickness=1,
        group_radius=0,
        group_text_vertical_offset=settings.group_text_vertical_offset,
        group_text_horizontal_offset=settings.group_text_horizontal_offset,
        marker_width=4,
        marker_projected_height=0,
        hidden_symbol_escape=("", ""),
        end_class=settings.end_class,
    )


_SYMBOL_TO_DIRECTION: dict[str, frozenset[str]] = {
    " ": frozenset(""),
    "↓": frozenset("ns"),
    "↑": frozenset("ns"),
    "→": frozenset("ew"),
    "←": frozenset("ew"),
    "─": frozenset("ew"),
    "━": frozenset("EW"),
    "│": frozenset("ns"),
    "┃": frozenset("NS"),
    "┌": frozenset("es"),
    "┍": frozenset("Es"),
    "┎": frozenset("eS"),
    "┏": frozenset("ES"),
    "┐": frozenset("sw"),
    "┑": frozenset("sW"),
    "┒": frozenset("Sw"),
    "┓": frozenset("SW"),
    "└": frozenset("ne"),
    "┕": frozenset("nE"),
    "┖": frozenset("Ne"),
    "┗": frozenset("NE"),
    "┘": frozenset("wn"),
    "┙": frozenset("Wn"),
    "┚": frozenset("wN"),
    "┛": frozenset("WN"),
    "├": frozenset("nes"),
    "┝": frozenset("nEs"),
    "┞": frozenset("Nes"),
    "┟": frozenset("neS"),
    "┠": frozenset("NeS"),
    "┡": frozenset("NEs"),
    "┢": frozenset("nES"),
    "┣": frozenset("NES"),
    "┤": frozenset("nsw"),
    "┥": frozenset("nsW"),
    "┦": frozenset("Nsw"),
    "┧": frozenset("nSw"),
    "┨": frozenset("NSw"),
    "┩": frozenset("NsW"),
    "┪": frozenset("nSW"),
    "┫": frozenset("NSW"),
    "┬": frozenset("esw"),
    "┭": frozenset("Esw"),
    "┮": frozenset("esW"),
    "┯": frozenset("EsW"),
    "┰": frozenset("eSw"),
    "┱": frozenset("ESw"),
    "┲": frozenset("eSW"),
    "┳": frozenset("ESW"),
    "┴": frozenset("new"),
    "┵": frozenset("neW"),
    "┶": frozenset("nEw"),
    "┷": frozenset("nEW"),
    "┸": frozenset("New"),
    "┹": frozenset("NeW"),
    "┺": frozenset("NEw"),
    "┻": frozenset("NEW"),
    "┼": frozenset("nesw"),
    "┽": frozenset("nEsw"),
    "┾": frozenset("nesW"),
    "┿": frozenset("nEsW"),
    "╀": frozenset("Nesw"),
    "╁": frozenset("neSw"),
    "╂": frozenset("NeSw"),
    "╃": frozenset("NesW"),
    "╄": frozenset("NEsw"),
    "╅": frozenset("neSW"),
    "╆": frozenset("nESw"),
    "╇": frozenset("NesW"),
    "╈": frozenset("nESW"),
    "╉": frozenset("NeSW"),
    "╊": frozenset("NESw"),
    "╋": frozenset("NESW"),
    "╴": frozenset("w"),
    "╵": frozenset("n"),
    "╶": frozenset("e"),
    "╷": frozenset("s"),
    "╸": frozenset("W"),
    "╹": frozenset("N"),
    "╺": frozenset("E"),
    "╻": frozenset("S"),
    "╼": frozenset("Ew"),
    "╽": frozenset("nS"),
    "╾": frozenset("eW"),
    "╿": frozenset("Ns"),
    "╭": frozenset("es"),
    "╮": frozenset("sw"),
    "╰": frozenset("ne"),
    "╯": frozenset("wn"),
}


_DIRECTION_TO_SYMBOL = {v: k for k, v in _SYMBOL_TO_DIRECTION.items()}

s = frozenset("NESW")
for k, v in list(_DIRECTION_TO_SYMBOL.items()):
    symbols = "".join(k & s).lower()
    for i in range(len(symbols)):
        for comb in itertools.combinations(symbols, i + 1):
            _DIRECTION_TO_SYMBOL[k | frozenset(comb)] = v
del s


class TextRender(Render[T], _t.Generic[T]):
    def __init__(
        self,
        width: int,
        height: int,
        settings: LayoutSettings[T],
        dump_debug_data: bool = False,
    ):
        super().__init__(settings, dump_debug_data)

        self._pos = Vec(0, 0)
        self._field = [[" "] * width for _ in range(height)]

    def write(self, f=sys.stdout):
        for line in self._field:
            for g in line:
                f.write(g)
            f.write("\n")
        f.flush()

    def to_string(self):
        stream = io.StringIO()
        self.write(stream)
        return stream.getvalue()

    def line(self, pos: Vec, reverse: bool = False, css_class: str = "") -> Line:
        return TextRender._TextLine(self, pos, reverse)

    def node(
        self,
        pos: Vec,
        style: NodeStyle,
        css_class: str | None,
        content_width: int,
        up: int,
        down: int,
        radius: int,
        padding: int,
        text_width: int,
        text_height: int,
        text: str,
        href: str | None,
        title: str | None,
    ):
        if style == NodeStyle.TERMINAL:
            ch = "┤├┌┐└┘─│"
        elif style == NodeStyle.NON_TERMINAL:
            ch = "╢╟╔╗╚╝═║"
        else:
            ch = "╴╶      "

        lines = text.splitlines()
        height = len(lines)
        offset_top = height // 2
        offset_bottom = height - offset_top

        for j, line in enumerate(lines):
            line_width, _ = _TEXT_MEASURE.measure(line)
            x = pos.x + padding
            y = pos.y - offset_top + j
            self._field[y][x] = line
            self._field[y][pos.x] = ch[7]
            self._field[y][pos.x + content_width - 1] = ch[7]
            for j in range(1, line_width):
                self._field[y][x + j] = ""

        self._field[pos.y - offset_top - 1][pos.x] = ch[2]
        for i in range(1, content_width):
            self._field[pos.y - offset_top - 1][pos.x + i] = ch[6]
        self._field[pos.y - offset_top - 1][pos.x + content_width - 1] = ch[3]

        self._field[pos.y + offset_bottom][pos.x] = ch[4]
        for i in range(1, content_width):
            self._field[pos.y + offset_bottom][pos.x + i] = ch[6]
        self._field[pos.y + offset_bottom][pos.x + content_width - 1] = ch[5]

        self._field[pos.y][pos.x] = ch[0]
        self._field[pos.y][pos.x + content_width - 1] = ch[1]

    def group(
        self,
        pos: Vec,
        width: int,
        height: int,
        css_class: str | None,
        text_width: int,
        text_height: int,
        text: str | None,
        href: str | None,
        title: str | None,
    ):
        self._write_cell(pos, "ES")
        self._write_cell(pos + Vec(width - 1, 0), "SW")
        self._write_cell(pos + Vec(0, height), "NE")
        self._write_cell(pos + Vec(width - 1, height), "NW")
        for x in range(pos.x + 1, pos.x + width - 1):
            self._write_cell(Vec(x, pos.y), "EW")
            self._write_cell(Vec(x, pos.y + height), "EW")
        for y in range(pos.y + 1, pos.y + height):
            self._write_cell(Vec(pos.x, y), "NS")
            self._write_cell(Vec(pos.x + width - 1, y), "NS")

        if not text:
            return

        lines = text.splitlines()
        text_pos = pos + Vec(
            self.settings.group_text_horizontal_offset + self.settings.group_thickness,
            -self.settings.group_text_vertical_offset - len(lines),
        )
        for i, line in enumerate(lines):
            line_width, _ = _TEXT_MEASURE.measure(line)
            y = text_pos.y + i
            self._field[y][text_pos.x] = line
            for j in range(1, line_width):
                self._field[y][text_pos.x + j] = ""
            if y == pos.y:
                if self.settings.group_text_horizontal_offset > 0:
                    self._field[y][text_pos.x - 1] = "╸"
                elif self.settings.group_text_horizontal_offset == 0:
                    self._field[y][text_pos.x - 1] = "╻"
                self._field[y][text_pos.x + line_width] = "╺"

    def left_marker(self, pos: Vec):
        w = self.settings.marker_width

        for i in range(w):
            self._field[pos.y][pos.x + i] = "─"

        if self.settings.end_class == EndClass.SIMPLE:
            self._field[pos.y][pos.x] = "├"
        elif self.settings.end_class == EndClass.COMPLEX:
            self._field[pos.y][pos.x] = "├"
            self._field[pos.y][pos.x + 1] = "┼"
        else:
            raise NotImplementedError(f"unknown end class {self.settings.end_class}")

    def right_marker(self, pos: Vec):
        w = self.settings.marker_width

        for i in range(w):
            self._field[pos.y][pos.x + i] = "─"

        if self.settings.end_class == EndClass.SIMPLE:
            self._field[pos.y][pos.x + w - 1] = "┤"
        elif self.settings.end_class == EndClass.COMPLEX:
            self._field[pos.y][pos.x + w - 1] = "┤"
            self._field[pos.y][pos.x + w - 2] = "┼"
        else:
            raise NotImplementedError(f"unknown end class {self.settings.end_class}")

    def _write_cell(self, pos: Vec, d: str):
        if self._field[pos.y][pos.x] not in _SYMBOL_TO_DIRECTION:
            pass
        elif self._field[pos.y][pos.x] == " ":
            self._field[pos.y][pos.x] = _DIRECTION_TO_SYMBOL[frozenset(d)]
        else:
            old_d = frozenset(d)
            new_d = old_d | _SYMBOL_TO_DIRECTION.get(
                self._field[pos.y][pos.x], frozenset()
            )
            if old_d != new_d:
                self._field[pos.y][pos.x] = _DIRECTION_TO_SYMBOL[new_d]

    class _TextLine(Line):
        def __init__(self, render: TextRender[_t.Any], pos: Vec, reverse: bool):
            self._render = render
            self._pos = Vec(pos.x, pos.y)
            self._reverse = reverse

        def segment_abs(
            self, x: int, arrow_begin: bool = False, arrow_end: bool = False
        ) -> Line:
            w = x - self._pos.x

            if w > 0:
                for i in range(w):
                    self._render._write_cell(self._pos + Vec(i, 0), "we")

                if arrow_begin:
                    self._render._field[self._pos.y][self._pos.x] = "→"

                self._pos.x += w

                if arrow_end:
                    self._render._field[self._pos.y][self._pos.x - 1] = "→"
            elif w < 0:
                for i in range(-1, w - 1, -1):
                    self._render._write_cell(self._pos + Vec(i, 0), "we")

                if arrow_begin:
                    self._render._field[self._pos.y][self._pos.x - 1] = "←"

                self._pos.x += w

                if arrow_end:
                    self._render._field[self._pos.y][self._pos.x] = "←"

            return self

        def bend(
            self,
            y: int,
            coming_from: Direction,
            coming_to: Direction | None,
            arrow_begin: bool = False,
            arrow_end: bool = False,
        ) -> Line:
            h = y - self._pos.y

            if coming_from == "e":
                self._pos.x -= 1

            s = "↓" if h > 0 else "↑"

            if arrow_begin:
                if h > 0:
                    self._render._field[self._pos.y + 1][self._pos.x] = s
                elif h < 0:
                    self._render._field[self._pos.y - 1][self._pos.x] = s

            if h == 0:
                self._render._write_cell(self._pos, coming_from + (coming_to or ""))
            elif h > 0:
                for i in range(1, h):
                    self._render._write_cell(self._pos + Vec(0, i), "ns")
                self._render._write_cell(self._pos, coming_from + "s")
                self._render._write_cell(self._pos + Vec(0, h), "n" + (coming_to or ""))
            else:
                for i in range(-1, h, -1):
                    self._render._write_cell(self._pos + Vec(0, i), "ns")
                self._render._write_cell(self._pos, coming_from + "n")
                self._render._write_cell(self._pos + Vec(0, h), "s" + (coming_to or ""))

            self._pos.y += h

            if arrow_end:
                if h > 0:
                    self._render._field[self._pos.y - 1][self._pos.x] = s
                elif h < 0:
                    self._render._field[self._pos.y + 1][self._pos.x] = s

            if coming_to == "e":
                self._pos.x += 1

            return self



from __future__ import annotations

import typing as _t

__all__ = [
    "HrefResolver",
]

T = _t.TypeVar("T")


class HrefResolver(_t.Generic[T]):
    """
    An interface that allows pre-processing all hyperlinks in nodes.

    """

    def resolve(
        self, text: str, href: str | None, title: str | None, resolver_data: T | None
    ):
        """
        Called before rendering a node.

        Should return a tuple ``(text, href, title)``, possibly modified.

        :param text:
            text of the node.
        :param href:
            hyperlink attached to the node.
        :param title:
            title for hyperlink.

        """

        return text, href, title



from dataclasses import dataclass

from syntax_diagrams._impl.vec import Vec


@dataclass(slots=True)
class RidgeLine:
    """
    Represents a ridge line.

    """

    before: int
    ridge: list[Vec]

    def __add__(self, rhs: Vec):
        return RidgeLine(self.before + rhs.y, [p + rhs for p in self.ridge])

    def __sub__(self, rhs: Vec):
        return RidgeLine(self.before - rhs.y, [p - rhs for p in self.ridge])


def merge_ridge_lines(lhs: RidgeLine, rhs: RidgeLine, cmp=max) -> RidgeLine:
    """
    Merge two ridge lines by taking the maximum height at each position.

    """

    before = cmp(lhs.before, rhs.before)

    l = lhs.ridge
    r = rhs.ridge
    result: list[Vec] = []
    i = j = 0
    current_l_height = lhs.before
    current_r_height = rhs.before

    while i < len(l) or j < len(r):
        if i >= len(l):
            x = r[j].x
        elif j >= len(r):
            x = l[i].x
        else:
            x = min(l[i].x, r[j].x)

        if i < len(l) and l[i].x == x:
            current_l_height = l[i].y
            i += 1
        if j < len(r) and r[j].x == x:
            current_r_height = r[j].y
            j += 1

        merged_height = cmp(current_l_height, current_r_height)

        if not result or result[-1].y != merged_height:
            result.append(Vec(x, merged_height))

    return RidgeLine(before, result)


def reverse_ridge_line(lhs: RidgeLine, pivot: int) -> RidgeLine:
    if not lhs.ridge:
        return lhs

    before = lhs.ridge[-1].y
    result: list[Vec] = []
    for i, p in enumerate(lhs.ridge):
        if i == 0:
            y = lhs.before
        else:
            y = lhs.ridge[i - 1].y
        result.append(Vec(pivot - p.x, y))
    result.reverse()

    return RidgeLine(before, result)


def find_distance(lhs: RidgeLine, rhs: RidgeLine) -> int:
    l = lhs.ridge
    r = rhs.ridge
    i = j = 0
    current_l_height = lhs.before
    current_r_height = rhs.before

    d = current_l_height + current_r_height

    while i < len(l) or j < len(r):
        if i >= len(l):
            x = r[j].x
        elif j >= len(r):
            x = l[i].x
        else:
            x = min(l[i].x, r[j].x)

        if i < len(l) and l[i].x == x:
            current_l_height = l[i].y
            i += 1
        if j < len(r) and r[j].x == x:
            current_r_height = r[j].y
            j += 1

        d = max(d, current_l_height + current_r_height)

    return d



from __future__ import annotations

import math
import re
import typing as _t
import unicodedata

import grapheme

from syntax_diagrams._impl.render import (
    LayoutContext,
    LayoutSettings,
    NodeStyle,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.tree import Element
from syntax_diagrams._impl.vec import Vec

T = _t.TypeVar("T")


class Node(Element[T], _t.Generic[T]):
    _style: NodeStyle
    _text: str
    _href: str | None
    _title: str | None
    _resolve: bool
    _resolver_data: T | None
    _css_class: str

    _horizontal_padding: int
    _radius: int
    _text_width: int
    _text_height: int
    _processed_text: str
    _processed_href: str | None
    _processed_title: str | None

    def __init__(
        self,
        render: NodeStyle,
        text: str,
        href: str | None = None,
        title: str | None = None,
        css_class: str | None = None,
        resolve: bool | None = None,
        resolver_data: T | None = None,
    ):
        self._style = render
        self._text = text
        self._href = href
        self._title = title
        self._resolve = resolve if resolve is not None else True
        self._resolver_data = resolver_data
        self._css_class = css_class or ""

    def _calculate_content_layout(
        self, settings: LayoutSettings[T], context: LayoutContext
    ):
        self._isolate()

        if self._resolve:
            self._processed_text, self._processed_href, self._processed_title = (
                settings.href_resolver.resolve(
                    self._text, self._href, self._title, self._resolver_data
                )
            )
        else:
            self._processed_text, self._processed_href, self._processed_title = (
                self._text,
                self._href,
                self._title,
            )

        self._processed_text = _reveal_hidden_symbols(
            self._processed_text, settings.hidden_symbol_escape
        )

        if self._style == NodeStyle.TERMINAL:
            text_measure = settings.terminal_text_measure
            self._horizontal_padding = settings.terminal_horizontal_padding
            self._vertical_padding = settings.terminal_vertical_padding
            self._radius = settings.terminal_radius
        elif self._style == NodeStyle.NON_TERMINAL:
            text_measure = settings.non_terminal_text_measure
            self._horizontal_padding = settings.non_terminal_horizontal_padding
            self._vertical_padding = settings.non_terminal_vertical_padding
            self._radius = settings.non_terminal_radius
        else:
            text_measure = settings.comment_text_measure
            self._horizontal_padding = settings.comment_horizontal_padding
            self._vertical_padding = settings.comment_vertical_padding
            self._radius = settings.comment_radius

        self._text_width, self._text_height = text_measure.measure(self._processed_text)

        self.display_width = self.content_width = (
            self._text_width + 2 * self._horizontal_padding
        )
        self.height = 0
        self.up = self.down = math.ceil(self._text_height / 2) + self._vertical_padding
        self.start_margin = self.end_margin = settings.horizontal_seq_separation

    def _render_content(self, render: Render[T], context: RenderContext):
        if not context.reverse:
            pos = context.pos
        else:
            pos = context.pos - Vec(self.content_width, 0)
        render.node(
            pos=pos,
            style=self._style,
            css_class=self._css_class,
            content_width=self.content_width,
            up=self.up,
            down=self.down,
            radius=self._radius,
            padding=self._horizontal_padding,
            text_width=self._text_width,
            text_height=self._text_height,
            text=self._processed_text,
            href=self._processed_href,
            title=self._processed_title,
        )

    def __str__(self):
        if re.match(r"^[a-zA-Z0-9_-]+$", self._text) is not None:
            return self._text
        else:
            return repr(self._text)


_CHAR_NAMES = {
    "\u0000": "<NUL>",
    "\u0001": "<SOH>",
    "\u0002": "<STX>",
    "\u0003": "<ETX>",
    "\u0004": "<EOT>",
    "\u0005": "<ENQ>",
    "\u0006": "<ACK>",
    "\u0007": "\\a",
    "\u0008": "\\b",
    "\u0009": "\\t",
    "\u000a": "\\n",
    "\u000b": "\\v",
    "\u000c": "\\f",
    "\u000d": "\\r",
    "\u000e": "<SO>",
    "\u000f": "<SI>",
    "\u0010": "<DLE>",
    "\u0011": "<DC1>",
    "\u0012": "<DC2>",
    "\u0013": "<DC3>",
    "\u0014": "<DC4>",
    "\u0015": "<NAK>",
    "\u0016": "<SYN>",
    "\u0017": "<ETB>",
    "\u0018": "<CAN>",
    "\u0019": "<EM>",
    "\u001a": "<SUB>",
    "\u001b": "<ESC>",
    "\u001c": "<FS>",
    "\u001d": "<GS>",
    "\u001e": "<RS>",
    "\u001f": "<US>",
    "\u007f": "<DEL>",
    "\u0080": "<PAD>",
    "\u0081": "<HOP>",
    "\u0082": "<BPH>",
    "\u0083": "<NBH>",
    "\u0084": "<IND>",
    "\u0085": "<NEL>",
    "\u0086": "<SSA>",
    "\u0087": "<ESA>",
    "\u0088": "<HTS>",
    "\u0089": "<HTJ>",
    "\u008a": "<VTS>",
    "\u008b": "<PLD>",
    "\u008c": "<PLU>",
    "\u008d": "<RI>",
    "\u008e": "<SS2>",
    "\u008f": "<SS3>",
    "\u0090": "<DCS>",
    "\u0091": "<PU1>",
    "\u0092": "<PU2>",
    "\u0093": "<STS>",
    "\u0094": "<CCH>",
    "\u0095": "<MW>",
    "\u0096": "<SPA>",
    "\u0097": "<EPA>",
    "\u0098": "<SOS>",
    "\u0099": "<SGCI>",
    "\u009a": "<SCI>",
    "\u009b": "<CSI>",
    "\u009c": "<ST>",
    "\u009d": "<OSC>",
    "\u009e": "<PM>",
    "\u009f": "<APC>",
    "\u00a0": "<NBSP>",
    "\u00ad": "<SHY>",
}


def _reveal_hidden_symbols(s: str, e: tuple[str, str], ignore: str = " ") -> str:
    res = ""
    for g in grapheme.graphemes(s):
        if len(g) > 1 and not g.isspace():
            res += g
            continue
        for c in g:
            if c in ignore:
                res += c
                continue
            if name := _CHAR_NAMES.get(c):
                res += f"{e[0]}{name}{e[1]}"
                continue
            cat = unicodedata.category(c)
            if cat[0] in "MCZ":
                name = unicodedata.name(c, None)
                if not name:
                    o = ord(c)
                    if o > 0xFFFF:
                        name = f"<U{o:08x}>"
                    else:
                        name = f"<U{o:04x}>"
                res += f"{e[0]}{name}{e[1]}"
            else:
                res += c

    return res



import syntax_diagrams as rr
from syntax_diagrams._impl.load import load
from syntax_diagrams._impl.render import (
    ConnectionType,
    LayoutContext,
    RenderContext,
)
from syntax_diagrams._impl.render.text import TextRender
from syntax_diagrams._impl.vec import Vec


def test_line(text_layout_settings):
    render = TextRender(5, 2, text_layout_settings)
    ((render).line(Vec(1, 0)).segment_abs(4))
    ((render).line(Vec(4, 1)).segment_abs(1))
    # fmt: off
    expected = (
        " ─── \n"
        " ─── \n"
    )
    # fmt: on
    assert render.to_string() == expected


def test_line_arrow(text_layout_settings):
    render = TextRender(5, 4, text_layout_settings)
    ((render).line(Vec(0, 0)).segment_abs(5, arrow_begin=True))
    ((render).line(Vec(0, 1)).segment_abs(5, arrow_end=True))
    ((render).line(Vec(5, 2)).segment_abs(0, arrow_begin=True))
    ((render).line(Vec(5, 3)).segment_abs(0, arrow_end=True))
    # fmt: off
    expected = (
        "→────\n"
        "────→\n"
        "────←\n"
        "←────\n"
    )
    # fmt: on
    assert render.to_string() == expected


def test_line_bend_reverse(text_layout_settings):
    render = TextRender(5, 11, text_layout_settings)
    (
        (render)
        .line(Vec(3, 0))
        .segment_abs(4)
        .bend_backward_reverse_abs(0)
        .segment_abs(1)
        .bend_backward_abs(0)
        .segment_abs(2)
    )
    (
        (render)
        .line(Vec(3, 1))
        .segment_abs(4)
        .bend_backward_reverse_abs(2)
        .segment_abs(1)
        .bend_backward_abs(1)
        .segment_abs(2)
    )
    (
        (render)
        .line(Vec(3, 4))
        .segment_abs(4)
        .bend_backward_reverse_abs(3)
        .segment_abs(1)
        .bend_backward_abs(4)
        .segment_abs(2)
    )
    (
        (render)
        .line(Vec(3, 5))
        .segment_abs(4)
        .bend_backward_reverse_abs(7)
        .segment_abs(1)
        .bend_backward_abs(5)
        .segment_abs(2)
    )
    (
        (render)
        .line(Vec(3, 10))
        .segment_abs(4)
        .bend_backward_reverse_abs(8)
        .segment_abs(1)
        .bend_backward_abs(10)
        .segment_abs(2)
    )
    # fmt: off
    expected = (
        "╶───╴\n"
        "╭─ ─╮\n"
        "╰───╯\n"
        "╭───╮\n"
        "╰─ ─╯\n"
        "╭─ ─╮\n"
        "│   │\n"
        "╰───╯\n"
        "╭───╮\n"
        "│   │\n"
        "╰─ ─╯\n"
    )
    # fmt: on
    assert render.to_string() == expected


def test_line_bend_forward(text_layout_settings):
    render = TextRender(5, 11, text_layout_settings)
    ((render).line(Vec(0, 0)).segment_abs(2).bend_forward_abs(0).segment_abs(5))
    ((render).line(Vec(0, 1)).segment_abs(2).bend_forward_abs(2).segment_abs(5))
    ((render).line(Vec(0, 4)).segment_abs(2).bend_forward_abs(3).segment_abs(5))
    ((render).line(Vec(0, 5)).segment_abs(2).bend_forward_abs(7).segment_abs(5))
    ((render).line(Vec(0, 10)).segment_abs(2).bend_forward_abs(8).segment_abs(5))
    # fmt: off
    expected = (
        "─────\n"
        "──╮  \n"
        "  ╰──\n"
        "  ╭──\n"
        "──╯  \n"
        "──╮  \n"
        "  │  \n"
        "  ╰──\n"
        "  ╭──\n"
        "  │  \n"
        "──╯  \n"
    )
    # fmt: on
    assert render.to_string() == expected


def test_node(text_layout_settings):
    render = TextRender(27, 9, text_layout_settings)

    node = load(rr.comment("fully automated"), lambda x: x)
    node.calculate_layout(text_layout_settings, LayoutContext(width=25, is_outer=True))
    node.render(
        render,
        RenderContext(
            pos=Vec(0, 1),
            reverse=False,
            start_connection_pos=Vec(0, 1),
            end_connection_pos=Vec(27, 1),
        ),
    )

    node = load(rr.terminal("luxury"), lambda x: x)
    node.calculate_layout(text_layout_settings, LayoutContext(width=25, is_outer=True))
    node.render(
        render,
        RenderContext(
            pos=Vec(0, 4),
            reverse=False,
            start_connection_pos=Vec(0, 4),
            end_connection_pos=Vec(27, 4),
        ),
    )

    node = load(rr.non_terminal("🌈gay space communism🌈"), lambda x: x)
    node.calculate_layout(text_layout_settings, LayoutContext(width=25, is_outer=True))
    node.render(
        render,
        RenderContext(
            pos=Vec(0, 7),
            reverse=False,
            start_connection_pos=Vec(0, 4),
            end_connection_pos=Vec(27, 4),
        ),
    )

    # fmt: off
    expected = (
        "                           \n"
        "╴ fully automated ╶────────\n"
        "                           \n"
        "┌────────┐                 \n"
        "┤ luxury ├─────────────────\n"
        "└────────┘                 \n"
        "╔═════════════════════════╗\n"
        "╢ 🌈gay space communism🌈 ╟\n"
        "╚═════════════════════════╝\n"
    )
    # fmt: on
    assert render.to_string() == expected


def test_node_enter(text_layout_settings):
    render = TextRender(27, 7, text_layout_settings)
    node = load(rr.terminal("XXX"), lambda x: x)
    node.calculate_layout(text_layout_settings, LayoutContext(width=25, is_outer=True))
    node.render(
        render,
        RenderContext(
            pos=Vec(0, 3),
            reverse=False,
            start_connection_pos=Vec(0, 3),
            end_connection_pos=Vec(27, 3),
        ),
    )
    # fmt: off
    expected = (
        "                           \n"
        "                           \n"
        "┌─────┐                    \n"
        "┤ XXX ├────────────────────\n"
        "└─────┘                    \n"
        "                           \n"
        "                           \n"
    )
    # fmt: on
    assert render.to_string() == expected

    render = TextRender(27, 7, text_layout_settings)
    node = load(rr.terminal("XXX"), lambda x: x)
    node.calculate_layout(
        text_layout_settings,
        LayoutContext(
            width=25,
            is_outer=True,
            start_connection=ConnectionType.SPLIT,
            end_connection=ConnectionType.SPLIT,
        ),
    )
    node.render(
        render,
        RenderContext(
            pos=Vec(0, 3),
            reverse=False,
            start_connection_pos=Vec(0, 0),
            end_connection_pos=Vec(27, 6),
        ),
    )
    # fmt: off
    expected = (
       "╮                          \n"
       "↓                          \n"
       "│ ┌─────┐                  \n"
       "╰→┤ XXX ├────────────────→╮\n"
       "  └─────┘                 │\n"
       "                          ↓\n"
       "                          ╰\n"
    )
    # fmt: on
    assert render.to_string() == expected

    render = TextRender(27, 7, text_layout_settings)
    node = load(rr.terminal("XXX"), lambda x: x)
    node.calculate_layout(
        text_layout_settings,
        LayoutContext(
            width=25,
            is_outer=True,
            start_connection=ConnectionType.STACK,
            end_connection=ConnectionType.STACK,
        ),
    )
    node.render(
        render,
        RenderContext(
            pos=Vec(0, 3),
            reverse=False,
            start_connection_pos=Vec(1, 0),
            end_connection_pos=Vec(26, 6),
        ),
    )
    # fmt: off
    expected = (
       "╭                          \n"
       "↓                          \n"
       "│ ┌─────┐                  \n"
       "╰→┤ XXX ├────────────────→╮\n"
       "  └─────┘                 │\n"
       "                          ↓\n"
       "                          ╯\n"
       ""
    )
    # fmt: on
    assert render.to_string() == expected



from __future__ import annotations

import traceback
import typing as _t

from flask import Flask, Response, request

import syntax_diagrams

app = Flask(__name__)

START_PAGE = """
<!DOCTYPE html>
<html>
<head>
    <title>Syntax Diagram Debug Viewer Backend</title>
</head>
<body>
    <h1>Syntax Diagram Debug Viewer Backend</h1>
</body>
</html>
"""


@app.route("/", methods=["GET", "POST"])
def index():
    if request.method == "POST":
        try:
            return render_diagram()
        except Exception:
            return dict(error=traceback.format_exc()), 500

    return START_PAGE


def render_diagram():
    data = request.get_json()
    try:
        diagram = data["diagram"]
        rawSettings = data["settings"]
        if rawSettings["render"] == "svg":
            # fmt: off
            settings = syntax_diagrams.SvgRenderSettings(
                max_width=rawSettings["svgMaxWidth"],
                reverse=rawSettings["reverse"],
                title=rawSettings["svgTitle"],
                description=rawSettings["svgDescription"],
                vertical_choice_separation_outer=rawSettings["svgVerticalChoiceSeparationOuter"],
                vertical_choice_separation=rawSettings["svgVerticalChoiceSeparation"],
                vertical_seq_separation_outer=rawSettings["svgVerticalSeqSeparationOuter"],
                vertical_seq_separation=rawSettings["svgVerticalSeqSeparation"],
                horizontal_seq_separation=rawSettings["svgHorizontalSeqSeparation"],
                end_class=syntax_diagrams.EndClass(rawSettings["endClass"]),
                arc_radius=rawSettings["svgArcRadius"],
                arc_margin=rawSettings["svgArcMargin"],
                arrow_style=syntax_diagrams.ArrowStyle(rawSettings["svgArrowStyle"]),
                arrow_length=rawSettings["svgArrowLength"],
                arrow_cross_length=rawSettings["svgArrowCrossLength"],
                terminal_horizontal_padding=rawSettings["svgTerminalHorizontalPadding"],
                terminal_vertical_padding=rawSettings["svgTerminalVerticalPadding"],
                terminal_radius=rawSettings["svgTerminalRadius"],
                non_terminal_horizontal_padding=rawSettings["svgNonTerminalHorizontalPadding"],
                non_terminal_vertical_padding=rawSettings["svgNonTerminalVerticalPadding"],
                non_terminal_radius=rawSettings["svgNonTerminalRadius"],
                comment_horizontal_padding=rawSettings["svgCommentHorizontalPadding"],
                comment_vertical_padding=rawSettings["svgCommentVerticalPadding"],
                comment_radius=rawSettings["svgCommentRadius"],
                group_vertical_padding=rawSettings["svgGroupVerticalPadding"],
                group_horizontal_padding=rawSettings["svgGroupHorizontalPadding"],
                group_vertical_margin=rawSettings["svgGroupVerticalMargin"],
                group_horizontal_margin=rawSettings["svgGroupHorizontalMargin"],
                group_radius=rawSettings["svgGroupRadius"],
                group_text_vertical_offset=rawSettings["svgGroupTextVerticalOffset"],
                group_text_horizontal_offset=rawSettings["svgGroupTextHorizontalOffset"],
                css_style=None,
                css_class="syntax-diagram",
            )
            # fmt: on
        else:
            # fmt: off
            settings = syntax_diagrams.TextRenderSettings(
                max_width=rawSettings["textMaxWidth"],
                reverse=rawSettings["reverse"],
                vertical_choice_separation_outer=rawSettings["textVerticalChoiceSeparationOuter"],
                vertical_choice_separation=rawSettings["textVerticalChoiceSeparation"],
                vertical_seq_separation_outer=rawSettings["textVerticalSeqSeparationOuter"],
                vertical_seq_separation=rawSettings["textVerticalSeqSeparation"],
                horizontal_seq_separation=rawSettings["textHorizontalSeqSeparation"],
                end_class=syntax_diagrams.EndClass(rawSettings["endClass"]),
                group_vertical_padding=rawSettings["textGroupVerticalPadding"],
                group_horizontal_padding=rawSettings["textGroupHorizontalPadding"],
                group_vertical_margin=rawSettings["textGroupVerticalMargin"],
                group_horizontal_margin=rawSettings["textGroupHorizontalMargin"],
                group_text_vertical_offset=rawSettings["textGroupTextVerticalOffset"],
                group_text_horizontal_offset=rawSettings["textGroupTextHorizontalOffset"],
            )
            # fmt: on
    except KeyError as e:
        return f"Unknown key {e}", 400

    try:
        if rawSettings["render"] == "svg":
            rendered = syntax_diagrams.render_svg(
                diagram,
                settings=_t.cast(syntax_diagrams.SvgRenderSettings, settings),
                _dump_debug_data=True,  # type: ignore
            )
        else:
            rendered = syntax_diagrams.render_text(
                diagram,
                settings=_t.cast(syntax_diagrams.TextRenderSettings, settings),
                _dump_debug_data=True,  # type: ignore
            )
    except syntax_diagrams.LoadingError as e:
        return str(e), 400

    return Response(rendered, content_type="application/json")


if __name__ == "__main__":
    app.run(debug=True, host="0.0.0.0", port=9011)



from syntax_diagrams._impl.ridge_line import (
    RidgeLine,
    find_distance,
    merge_ridge_lines,
    reverse_ridge_line,
)
from syntax_diagrams._impl.vec import Vec


def test_empty_lines():
    empty = RidgeLine(0, [])
    line = RidgeLine(0, [Vec(0, 10), Vec(5, 20)])

    assert merge_ridge_lines(empty, line) == line
    assert merge_ridge_lines(line, empty) == line
    assert merge_ridge_lines(empty, empty) == empty


def test_single_point_lines():
    line1 = RidgeLine(0, [Vec(5, 10)])
    line2 = RidgeLine(0, [Vec(5, 20)])
    line3 = RidgeLine(0, [Vec(10, 15)])

    # Same x, different heights
    assert merge_ridge_lines(line1, line2) == RidgeLine(0, [Vec(5, 20)])

    # Different x positions
    assert merge_ridge_lines(line1, line3) == RidgeLine(0, [Vec(5, 10), Vec(10, 15)])


def test_non_overlapping_lines():
    left = RidgeLine(0, [Vec(0, 10), Vec(2, 15)])
    right = RidgeLine(0, [Vec(5, 20), Vec(7, 25)])

    expected = RidgeLine(0, [Vec(0, 10), Vec(2, 15), Vec(5, 20), Vec(7, 25)])
    assert merge_ridge_lines(left, right) == expected


def test_overlapping_lines_merge():
    line1 = RidgeLine(0, [Vec(0, 10), Vec(5, 20), Vec(10, 15)])
    line2 = RidgeLine(5, [Vec(2, 5), Vec(6, 25), Vec(8, 30)])

    expected = RidgeLine(5, [Vec(0, 10), Vec(5, 20), Vec(6, 25), Vec(8, 30)])
    result = merge_ridge_lines(line1, line2)
    assert result == expected


def test_overlapping_lines_step():
    line1 = RidgeLine(0, [Vec(0, 10), Vec(5, 20), Vec(8, 5)])
    line2 = RidgeLine(0, [Vec(2, 5), Vec(6, 10), Vec(10, 30)])

    expected = RidgeLine(0, [Vec(0, 10), Vec(5, 20), Vec(8, 10), Vec(10, 30)])
    result = merge_ridge_lines(line1, line2)
    assert result == expected


def test_identical_lines():
    line = RidgeLine(0, [Vec(0, 10), Vec(5, 20), Vec(10, 15)])

    assert merge_ridge_lines(line, line) == line


def test_same_x_different_heights():
    line1 = RidgeLine(0, [Vec(0, 10), Vec(5, 20), Vec(10, 30)])
    line2 = RidgeLine(0, [Vec(0, 15), Vec(5, 10), Vec(10, 25)])

    assert merge_ridge_lines(line1, line2) == RidgeLine(
        0, [Vec(0, 15), Vec(5, 20), Vec(10, 30)]
    )


def test_reverse_ridge_line_empty():
    empty = RidgeLine(0, [])

    expected = RidgeLine(0, [])
    assert reverse_ridge_line(empty, 0) == expected


def test_reverse_ridge_line_single_point():
    empty = RidgeLine(5, [Vec(10, 0)])

    expected = RidgeLine(0, [Vec(-10, 5)])
    assert reverse_ridge_line(empty, 0) == expected


def test_reverse_ridge_line():
    line = RidgeLine(0, [Vec(0, 10), Vec(5, 20), Vec(10, 15)])

    expected = RidgeLine(15, [Vec(-10, 20), Vec(-5, 10), Vec(0, 0)])
    assert reverse_ridge_line(line, 0) == expected


def test_reverse_ridge_line_pivot():
    line = RidgeLine(0, [Vec(0, 10), Vec(5, 20), Vec(10, 15)])

    expected = RidgeLine(15, [Vec(0, 20), Vec(5, 10), Vec(10, 0)])
    assert reverse_ridge_line(line, 10) == expected


def test_find_distance_empty():
    line1 = RidgeLine(5, [])
    line2 = RidgeLine(10, [])

    assert find_distance(line1, line2) == 15


def test_find_distance_step_up():
    line1 = RidgeLine(5, [Vec(5, 10)])
    line2 = RidgeLine(10, [])

    assert find_distance(line1, line2) == 20


def test_find_distance_step_down():
    line1 = RidgeLine(5, [Vec(5, 0)])
    line2 = RidgeLine(10, [])

    assert find_distance(line1, line2) == 15


def test_find_distance_single_point():
    line1 = RidgeLine(5, [Vec(5, 0)])
    line2 = RidgeLine(10, [Vec(5, 25)])

    assert find_distance(line1, line2) == 25


def test_find_distance_overlapping_lines():
    line1 = RidgeLine(5, [Vec(5, 0), Vec(10, -5), Vec(15, 15)])
    line2 = RidgeLine(0, [Vec(3, 5), Vec(5, 0), Vec(15, -6)])

    assert find_distance(line1, line2) == 10



from dataclasses import replace

import syntax_diagrams as rr
from syntax_diagrams._impl.load import load
from syntax_diagrams._impl.render import (
    ConnectionType,
    LayoutContext,
    RenderContext,
)
from syntax_diagrams._impl.render.svg import SvgRender
from syntax_diagrams._impl.vec import Vec
from syntax_diagrams.render import ArrowStyle


def test_line(regression, svg_layout_settings, svg_css):
    render = SvgRender(
        50, 15, svg_layout_settings, "", svg_css, None, None, ArrowStyle.NONE, 0, 0
    )
    ((render).line(Vec(10, 5)).segment_abs(40))
    ((render).line(Vec(40, 10)).segment_abs(10))
    regression(render.to_string())


def test_line_bend_reverse(regression, svg_layout_settings, svg_css):
    render = SvgRender(
        50, 210, svg_layout_settings, "", svg_css, None, None, ArrowStyle.NONE, 0, 0
    )
    (
        (render)
        .line(Vec(30, 10))
        .segment_abs(40)
        .bend_backward_reverse_abs(10)
        .segment_abs(10)
        .bend_backward_abs(10)
        .segment_abs(20)
    )
    (
        (render)
        .line(Vec(30, 30))
        .segment_abs(40)
        .bend_backward_reverse_abs(45)
        .segment_abs(10)
        .bend_backward_abs(30)
        .segment_abs(20)
    )
    (
        (render)
        .line(Vec(30, 80))
        .segment_abs(40)
        .bend_backward_reverse_abs(65)
        .segment_abs(10)
        .bend_backward_abs(80)
        .segment_abs(20)
    )
    (
        (render)
        .line(Vec(30, 100))
        .segment_abs(40)
        .bend_backward_reverse_abs(140)
        .segment_abs(10)
        .bend_backward_abs(100)
        .segment_abs(20)
    )
    (
        (render)
        .line(Vec(30, 200))
        .segment_abs(40)
        .bend_backward_reverse_abs(160)
        .segment_abs(10)
        .bend_backward_abs(200)
        .segment_abs(20)
    )
    regression(render.to_string())


def test_line_bend_forward(regression, svg_layout_settings, svg_css):
    render = SvgRender(
        50, 210, svg_layout_settings, "", svg_css, None, None, ArrowStyle.NONE, 0, 0
    )
    ((render).line(Vec(0, 10)).segment_abs(20).bend_forward_abs(10).segment_abs(50))
    ((render).line(Vec(0, 30)).segment_abs(20).bend_forward_abs(45).segment_abs(50))
    ((render).line(Vec(0, 80)).segment_abs(20).bend_forward_abs(65).segment_abs(50))
    ((render).line(Vec(0, 100)).segment_abs(20).bend_forward_abs(140).segment_abs(50))
    ((render).line(Vec(0, 200)).segment_abs(20).bend_forward_abs(160).segment_abs(50))
    regression(render.to_string())


def test_line_bend_forward_rtl(regression, svg_layout_settings, svg_css):
    render = SvgRender(
        50, 210, svg_layout_settings, "", svg_css, None, None, ArrowStyle.NONE, 0, 0
    )
    (
        (render)
        .line(Vec(50, 10), True)
        .segment_abs(30)
        .bend_forward_abs(10)
        .segment_abs(0)
    )
    (
        (render)
        .line(Vec(50, 30), True)
        .segment_abs(30)
        .bend_forward_abs(45)
        .segment_abs(0)
    )
    (
        (render)
        .line(Vec(50, 80), True)
        .segment_abs(30)
        .bend_forward_abs(65)
        .segment_abs(0)
    )
    (
        (render)
        .line(Vec(50, 100), True)
        .segment_abs(30)
        .bend_forward_abs(140)
        .segment_abs(0)
    )
    (
        (render)
        .line(Vec(50, 200), True)
        .segment_abs(30)
        .bend_forward_abs(160)
        .segment_abs(0)
    )
    regression(render.to_string())


def test_node(regression, svg_layout_settings, svg_css):
    render = SvgRender(
        200, 150, svg_layout_settings, "", svg_css, None, None, ArrowStyle.NONE, 0, 0
    )

    node = load(rr.comment("fully automated"), lambda x: x)
    node.calculate_layout(svg_layout_settings, LayoutContext(width=200, is_outer=True))
    node.render(
        render,
        RenderContext(
            pos=Vec(5, 20),
            reverse=False,
            start_connection_pos=Vec(0, 20),
            end_connection_pos=Vec(200, 20),
        ),
    )

    node = load(rr.terminal("luxury"), lambda x: x)
    node.calculate_layout(svg_layout_settings, LayoutContext(width=200, is_outer=True))
    node.render(
        render,
        RenderContext(
            pos=Vec(5, 60),
            reverse=False,
            start_connection_pos=Vec(0, 60),
            end_connection_pos=Vec(200, 60),
        ),
    )

    node = load(rr.non_terminal("gay space communism"), lambda x: x)
    node.calculate_layout(svg_layout_settings, LayoutContext(width=200, is_outer=True))
    node.render(
        render,
        RenderContext(
            pos=Vec(5, 100),
            reverse=False,
            start_connection_pos=Vec(0, 100),
            end_connection_pos=Vec(200, 100),
        ),
    )

    regression(render.to_string())


def test_node_enter(regression, svg_layout_settings, svg_css):
    render = SvgRender(
        200, 100, svg_layout_settings, "", svg_css, None, None, ArrowStyle.NONE, 0, 0
    )
    node = load(rr.terminal("XXX"), lambda x: x)
    node.calculate_layout(svg_layout_settings, LayoutContext(width=200, is_outer=True))
    node.render(
        render,
        RenderContext(
            pos=Vec(10, 50),
            reverse=False,
            start_connection_pos=Vec(0, 50),
            end_connection_pos=Vec(200, 50),
        ),
    )
    regression(render.to_string(), "normal")

    render = SvgRender(
        200, 100, svg_layout_settings, "", svg_css, None, None, ArrowStyle.NONE, 0, 0
    )
    node = load(rr.terminal("XXX"), lambda x: x)
    node.calculate_layout(
        svg_layout_settings,
        LayoutContext(
            width=200,
            is_outer=True,
            start_connection=ConnectionType.SPLIT,
            end_connection=ConnectionType.SPLIT,
        ),
    )
    node.render(
        render,
        RenderContext(
            pos=Vec(10, 50),
            reverse=False,
            start_connection_pos=Vec(0, 0),
            end_connection_pos=Vec(200, 100),
        ),
    )
    regression(render.to_string(), "split")

    render = SvgRender(
        200, 100, svg_layout_settings, "", svg_css, None, None, ArrowStyle.NONE, 0, 0
    )
    node = load(rr.terminal("XXX"), lambda x: x)
    node.calculate_layout(
        svg_layout_settings,
        LayoutContext(
            width=200,
            is_outer=True,
            start_connection=ConnectionType.STACK,
            end_connection=ConnectionType.STACK,
        ),
    )
    node.render(
        render,
        RenderContext(
            pos=Vec(10, 50),
            reverse=False,
            start_connection_pos=Vec(10, 0),
            end_connection_pos=Vec(190, 100),
        ),
    )
    regression(render.to_string(), "stack")


def test_end_class(regression, svg_render_settings):
    rendered = rr.render_svg(
        rr.skip(),
        max_width=400,
        settings=replace(svg_render_settings, end_class=rr.EndClass.SIMPLE),
    )
    regression(rendered, "simple")

    rendered = rr.render_svg(
        rr.skip(),
        max_width=400,
        settings=replace(svg_render_settings, end_class=rr.EndClass.COMPLEX),
    )
    regression(rendered, "complex")


class TestArcRadius:
    def test_r10(self, regression, svg_render_settings):
        rendered = rr.render_svg(
            rr.choice(
                rr.terminal("A"),
                rr.terminal("B"),
            ),
            max_width=400,
            settings=replace(svg_render_settings, arc_radius=10),
        )
        regression(rendered)

    def test_r5(self, regression, svg_render_settings):
        rendered = rr.render_svg(
            rr.choice(
                rr.terminal("A"),
                rr.terminal("B"),
            ),
            max_width=400,
            settings=replace(svg_render_settings, arc_radius=5),
        )
        regression(rendered)

    def test_r0(self, regression, svg_render_settings):
        rendered = rr.render_svg(
            rr.choice(
                rr.terminal("A"),
                rr.terminal("B"),
            ),
            max_width=400,
            settings=replace(svg_render_settings, arc_radius=0),
        )
        regression(rendered)

    def test_r0_no_margin(self, regression, svg_render_settings):
        rendered = rr.render_svg(
            rr.choice(
                rr.terminal("A"),
                rr.terminal("B"),
            ),
            max_width=400,
            settings=replace(svg_render_settings, arc_radius=0, arc_margin=0),
        )
        regression(rendered)


class TestSeq:
    def test_break(self, regression, svg_render_settings):
        rendered = rr.render_svg(
            rr.sequence(
                rr.terminal("A"),
                rr.terminal("B"),
                rr.terminal("C"),
                rr.terminal("D"),
            ),
            max_width=200,
            settings=svg_render_settings,
        )
        regression(rendered)

    def test_soft_break(self, regression, svg_render_settings):
        rendered = rr.render_svg(
            rr.sequence(
                rr.terminal("A"),
                rr.terminal("B"),
                rr.terminal("C"),
                rr.terminal("D"),
                linebreaks=[
                    rr.LineBreak.SOFT,
                    rr.LineBreak.NO_BREAK,
                    rr.LineBreak.NO_BREAK,
                ],
            ),
            max_width=200,
            settings=svg_render_settings,
        )
        regression(rendered)

    def test_hard_break(self, regression, svg_render_settings):
        rendered = rr.render_svg(
            rr.sequence(
                rr.terminal("A"),
                rr.terminal("B"),
                rr.terminal("C"),
                rr.terminal("D"),
                linebreaks=[
                    rr.LineBreak.NO_BREAK,
                    rr.LineBreak.HARD,
                    rr.LineBreak.NO_BREAK,
                ],
            ),
            max_width=400,
            settings=svg_render_settings,
        )
        regression(rendered)



from __future__ import annotations

import math
import typing as _t

from syntax_diagrams._impl.render import (
    ConnectionType,
    LayoutContext,
    LayoutSettings,
    Render,
    RenderContext,
)
from syntax_diagrams._impl.tree import Element
from syntax_diagrams._impl.vec import Vec

T = _t.TypeVar("T")


class Skip(Element[T], _t.Generic[T]):
    def _calculate_content_layout(
        self, settings: LayoutSettings[T], context: LayoutContext
    ):
        self.__start_connection = context.start_connection
        self.__start_arc_size = self.__start_connection.arc_size(settings)
        self.__end_connection = context.end_connection
        self.__end_arc_size = self.__end_connection.arc_size(settings)

        self._isolate()

    def render(self, render: Render[T], context: RenderContext):
        render.enter(self)
        render.debug_ridge_line(context.pos, self, context.reverse)

        start_content_pos = context.pos + Vec(context.dir * self.__start_arc_size, 0)
        match self.__start_connection:
            case (
                ConnectionType.NULL | ConnectionType.NORMAL | ConnectionType.STACK_BOUND
            ):
                line = (render).line(
                    context.start_connection_pos,
                    context.reverse,
                    "dbg-isolated-line",
                )
            case ConnectionType.STACK:
                line = (
                    (render)
                    .line(
                        context.start_connection_pos,
                        context.reverse,
                        "dbg-isolated-line",
                    )
                    .bend_backward_abs(start_content_pos.y, arrow_begin=True)
                )
            case ConnectionType.SPLIT:
                line = (
                    (render)
                    .line(
                        context.start_connection_pos,
                        context.reverse,
                        "dbg-isolated-line",
                    )
                    .bend_forward_abs(start_content_pos.y, arrow_begin=True)
                )

        match self.__end_connection:
            case ConnectionType.NULL:
                pass
            case ConnectionType.NORMAL | ConnectionType.STACK_BOUND:
                (
                    (line).segment_abs(
                        context.end_connection_pos.x, arrow_begin=True, arrow_end=True
                    )
                )
            case ConnectionType.STACK:
                (
                    (line)
                    .segment_abs(
                        context.end_connection_pos.x, arrow_begin=True, arrow_end=True
                    )
                    .bend_backward_reverse_abs(
                        context.end_connection_pos.y, arrow_end=True
                    )
                )
            case ConnectionType.SPLIT:
                arc_radius = context.dir * math.ceil(2 * render.settings.arc_radius)
                (
                    (line)
                    .segment_abs(
                        context.end_connection_pos.x - arc_radius,
                        arrow_begin=True,
                        arrow_end=True,
                    )
                    .bend_forward_abs(context.end_connection_pos.y, arrow_end=True)
                )

        render.debug_pos(context.pos, "dbg-primary-pos")
        render.debug_pos(context.start_connection_pos, "dbg-primary-pos")
        render.debug_pos(context.end_connection_pos, "dbg-primary-pos")
        render.exit()

    def _render_content(self, render: Render[T], context: RenderContext):
        pass

    def __str__(self) -> str:
        return "<SKIP>"
